import re
import itertools
from collections import deque
import heapq

def chunkit(it,size):
    cur_chunk = []
    for el in it:
        cur_chunk.append(el)
        if len(cur_chunk)>=size:
            yield cur_chunk
            cur_chunk = []
    yield cur_chunk
    return
    
def grouper(it,n,fill=None):
    args = [it]*n
    return itertools.zip_longest(*args,fillvalue=fill)

class zip_variants():
    class Exhasted(Exception):
        pass

    def __init__(self,left,right,direction='both'):
        self.deque = [deque(),deque()]
        self.cache = deque()
        _left = zip(itertools.repeat(0),left)
        _right = zip(itertools.repeat(1),right)
        self.merged = heapq.merge(_left,_right,key=lambda i:i[1].start)

    def __iter__(self):
        return self

    def _trim_deques(self,val):
        for deque in self.deque:
            if not deque:
                continue
            while True:
                if deque and deque[0].end<val.start:
                    deque.popleft()
                else:
                    break

    def _next_(self):
        try:
            deque,val = self.cache.popleft()
        except IndexError:
            try:
                deque,val = next(self.merged)
                self.deque[deque].append(val)
            except StopIteration:
                raise self.Exhasted
        self._trim_deques(val)
        return deque,val

    def __next__(self):
        try:
            deque,vrt = self._next_()
        except self.Exhasted:
            raise StopIteration

        # ~deque+2 is merely a fast way to convert 0 to 1 and 1 to 0
        if self.deque[~deque+2] and \
                      self.deque[~deque+2][-1].start>vrt.end:
            v = (deque,vrt,list(self.deque[~deque+2]))
            return v
        else:
            while True:
                try:
                    deque2,vrt2 = next(self.merged)
                    self.deque[deque2].append(vrt2)
                except StopIteration:
                    return (deque,vrt,list(self.deque[~deque+2]))
                    break

                self.cache.append((deque2,vrt2))
                if vrt2.start>vrt.end:
                    v = (deque,vrt,list(self.deque[~deque+2]))
                    return v
                    break
                else:
                    continue

def _strip_ref_alt(ref,alt):
    # Check identical in the beginning
    i = 0
    mn = min(len(ref),len(alt))
    while i < mn:
        if ref[i] == alt[i]:
            i += 1
            continue
        else:
            break
    shift=i

    # Strip positions identical at the end
    j=1
    while j <= mn-i:
        if ref[-j] == alt[-j]:
            j += 1
            continue
        else:
            break
    j -= 1
    ref = ref[i:len(ref)-j]
    alt = alt[i:len(alt)-j]
    return ref,alt,shift

def rgn_from(s):
    m = re.match('(.+):([0-9,]+)-([0-9,]+)',s)
    if m:
        chrom,start,end = m.groups()
        return chrom,int(start.replace(',','')),\
            int(end.replace(',',''))
    else:
        raise ValueError('Could not parse region: {}'.format(s))

def no_ovlp(it):
    for chrom,chrom_it in itertools.groupby(it,lambda v: v.chrom):
        cur_chunk = []
        for vrt in chrom_it:
            if len(cur_chunk)>0:
                if vrt.start < max([i.end for i in cur_chunk]):
                    cur_chunk.append(vrt)
                else:
                    yield cur_chunk
                    cur_chunk = [vrt]
            else:
                cur_chunk = [vrt]
        if len(cur_chunk)>0:
            yield cur_chunk
        



"""
Module ``varset`` defines variant set classes which are containers of variants
supporting set-like operations, searching variants, export to NumPy...

There are several classes depending on whether all the variants are loaded
into memory or not and whether the object is mutable:

    - :class:`~genomvar.varset.VariantSet` is immutable, in-memory, random
      access supporting class

    - :class:`~genomvar.varset.MutableVariantSet` mutable, in-memory, random
      access

    - :class:`~genomvar.varset.VariantFileSet` immutable, VCF file storage, no
      random access

    - :class:`~genomvar.varset.IndexedVariantFileSet` immutable, VCF file
      storage, random access
"""
import warnings
from copy import copy
import itertools
from functools import lru_cache
import collections
import numpy as np
from rbi_tree.tree import ITree
from genomvar import variant, variantops as ops, MAX_END
from genomvar.variant import VariantBase,AmbigIndel,\
    GenomVariant,Haplotype,VariantFactory
from genomvar.utils import rgn_from,zip_variants,\
    chunkit,no_ovlp
from genomvar.vcf import (_get_reader, VCFReader, VCFWriter,
                          dtype0,dtype1, )
from genomvar.vcf_utils import (header as vcf_header,
                                row_tmpl, _make_field_writer_func,
                                field_writer_simple,
                                dtype2string, string2dtype)
from genomvar import OverlappingHaplotypeVars,\
    Reference,DifferentlySortedChromsError,\
    DuplicateVariants,SINGLETON,ChromSet

class VariantSetBase(object):
    """Base class for variant set subclasses.

    Attributes
    ----------
    reference: :class:`genomvar.Reference`
        Object representing reference
    chroms: :class:`genomvar.ChromSet`
        list-like representation of chromosomes in the set
    ctg_len: dict
        dict mapping chromosome names to corresponding lengths.
        
        This keeps track of the rightmost position of variants in the set
        if reference was not provided. If reference was given ``ctg_len`` 
        is taken from length of chromosomes in reference.
    """


    def __init__(self,variants=None,data=None,reference=None):
        if not reference is None:
            if isinstance(reference,Reference):
                self.reference=reference
            elif isinstance(reference,str):
                self.reference = Reference(reference,cache_dst=100000)
            else:
                raise TypeError('reference not understood')
            self.ctg_len = self.reference.ctg_len
            self._chroms = ChromSet(self.reference.get_chroms())
        else:
            self.reference = None
            self.ctg_len = {}
            self._chroms = ChromSet()
        
        self._data = data if data else {}
        self._factory = VariantFactory(reference=self.reference,
                                       normindel=False)
        self._fac = {'nonorm':self._factory}
        
    def get_chroms(self):
        return self.chroms
    
    def _find_iv(self,chrom,start=0,end=MAX_END):
        if chrom not in self._data:
            return iter([])
        for ivl in self._data[chrom].find(start,end):
            yield ivl

    def find_vrt(self,chrom=None,start=0,end=MAX_END,rgn=None,
                 expand=False,check_order=False):
        """Finds variants in specified region.

        If not ``chrom`` nor ``rgn`` parameters are given it 
        traverses all variants.

        Parameters
        ----------

        chrom: str, optional
            Chromosome to search variants on. If not given, all chroms
            are traversed and ``start`` and ``end`` parameters ignored.
        start: int, optional
            Start of search region (0-based). *Default: 0*
        end: int, optional
            End of search region (position ``end`` is excluded). *Default: 
            max chrom length*
        rgn: str, optional
            If given chrom, start and end parameters are
            ignored.  A string specifying search region, e.g.
            ``chr15:365,008-365,958``.
        expand: bool, optional
            Indicates whether to traverse haplotypes. *Default: False*

        Returns
        -------
        List of :class:`genomvar.variant.VariantBase` subclass objects
        """
        # here only need to parse arguments correctly and delegate
        # actual job to _find_vrt
        if not rgn is None:
            chrom,start,end = rgn_from(rgn)
            for vrt in self._find_vrt(chrom,start,end,
                                       expand=expand):
                yield vrt
        elif not chrom is None:
            for vrt in self._find_vrt(chrom,start=start,end=end,
                                       expand=expand):
                yield vrt
        else:
            for b in self.iter_vrt(expand=expand):
                yield b

    def _find_chrom_vrt(self,chrom,expand=False):
        for vrt in self._find_vrt(chrom,expand=expand):
            yield vrt

    def iter_vrt_by_chrom(self,**kwds):
        """Iterates variants grouped by chromosome."""
        for chrom in self.chroms:
            yield (chrom,self._find_chrom_vrt(chrom))

    def iter_vrt(self,expand=False):
        """
        Iterate over all variants in the set.
        
        Parameters
        ----------
        expand : bool, optional
            If ``True`` variants in encountered haplotypes are yielded.
            *Default: False*
        Yields
        -------
        :class:`genomvar.variant.GenomVariant`
        """
        for chrom in sorted(self.chroms):
            for vrt in self._find_chrom_vrt(chrom,expand=expand):
                yield vrt

    def ovlp(self,vrt0,match_ambig=False):
        """
        Returns variants altering the same positions as ``vrt0``.
        
        Parameters
        ----------

        vrt0 : VariantBase-like
            Variant to search overlap with.
        match_ambig : bool
            If False ambigous indels treated as regular indels. *Default: False*

        Returns
        -------
        list of variants
        """
        chrom = vrt0.chrom
        ovlp = []
        if not match_ambig and vrt0.is_instance(variant.AmbigIndel):
            start = vrt0.act_start
            end = vrt0.act_end
        else:
            start = vrt0.start
            end = vrt0.end

        # Insertions need special treatment due to arbitrariness of
        # position definition
        if vrt0.is_instance(variant.Ins):
            for _vrt in self._find_vrt(chrom,start,end):
                if _vrt.is_instance(variant.AmbigIndel) and not match_ambig:
                    _vrt_end = _vrt.act_end
                    _vrt_start = _vrt.act_start
                else:
                    _vrt_end = _vrt.end
                    _vrt_start = _vrt.start
                # insertions in the same spot do overlap
                # if _vrt.is_instance(variant.Ins) and _vrt_start==start:
                #     ovlp.append(_vrt)
                #     continue
                # next is no overlap case
                if max(_vrt_start,start)>=min(_vrt_end,end):
                    continue
                # next is for case like
                # CT|TTCTTG
                #   T
                #    G
                # but if it is insertion than can add to overlap
                # due to identical position
                elif end<=_vrt_start+1 and not _vrt.is_instance(variant.Ins):
                    continue
                else:
                    ovlp.append(_vrt)
        else:
            ovlp += self._ovlp1(chrom,start,end,
                                match_ambig=match_ambig)

        return ovlp

    def _ovlp1(self,chrom,start,end,match_ambig=False):
        """Returns variants in the specified region. Locations of 
        indels are taken care of."""
        ovlp = []
        
        for _vrt in self._find_vrt(chrom,start,end):
            if _vrt.is_instance(variant.AmbigIndel) and not match_ambig:
                _vrt_end = _vrt.act_end
                _vrt_start = _vrt.act_start
            else:
                _vrt_end = _vrt.end
                _vrt_start = _vrt.start
            # if _vrt is insertion and the search region is further
            if _vrt.is_instance(variant.Ins) and _vrt_end<=start+1:
                continue
            # check maybe no overlp after ambiguity removal
            if max(_vrt_start,start)>=min(_vrt_end,end):
                continue
            else:
                ovlp.append(_vrt)
        return ovlp

    def match(self,vrt,match_partial=True,match_ambig=False):
        """
        Find variants mathing vrt.
        
        The method checks whether variant ``vrt``
        is present in the set.

        Parameters
        ----------
        vrt : VariantBase-like
            Variant to search match for. 
        match_partial: bool, optional
            If ``False`` common positions won't be searched in non-identical
            MNPs (relevant if ``vrt`` is MNP or SNP or haplotype, containing
            these). *Default: True*
        match_ambig: bool, optional
            If ``False`` ambigous indels will be treated as regular indels
            (relevant if ``vrt`` is Ambigindel or haplotype, containing ones).
            *Default: False*
        Returns
        -------
        list or dict with matching variants
            dict is returned only of ``vrt`` is a Haplotype

        """
        ovlp = list(self._find_vrt(vrt.chrom,vrt.start,vrt.end))
        return ops.matchv(vrt,ovlp,match_ambig=match_ambig,
                          match_partial=match_partial)

    def diff_vrt(self,other,match_partial=True,match_ambig=False):
        """
        Generate variants different between ``self`` and ``other``.
        
        This creates a lazily evaluated object of class :class:`CmpSet`. This
        object can be further inspected with ``.iter_vrt()`` or ``.region()``

        Parameters
        ----------
        other : variant set
            Variant set to compare against ``self``

        match_partial: bool, optional
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*

        match_ambig: bool, optional
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*
        Returns
        -------
        :class:`CmpSet`
            Transient comparison set
        """
        return self._cmp_vrt(other,action='diff',match_ambig=match_ambig,
                          match_partial=match_partial)

    def comm_vrt(self,other,match_partial=True,match_ambig=False):
        """
        Generate variants common between ``self`` and ``other``.
        
        This creates a lazily evaluated object of class :class:`CmpSet`. This
        object can be further inspected with ``.iter_vrt()`` or ``.region()``

        Parameters
        ----------
        other : variant set
            Variant set to compare against ``self``
        match_partial: bool, optional
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*
        match_ambig: bool, optional
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*
        Returns
        -------
        :class:`CmpSet`
            Transient comparison set
        """
        return self._cmp_vrt(other,action='comm',match_ambig=match_ambig,
                          match_partial=match_partial)


    def _cmp_vrt(self,other,action,match_partial=True,match_ambig=False):
        return CmpSet(self,other,action,match_partial=match_partial,
                      match_ambig=match_ambig)
    
    def to_vcf(self, out, reference=None, info_spec=None):
        """Writes a minimal VCF with variants from the set. 

        INFO and SAMPLE data from source data is not preserved.

        Parameters
        ----------

        out : handle or str
           If string then it's path to file, otherwise a handle 
           to write variants.

        reference : Reference or str, optional
           If string then it's path to reference FASTA. Otherwise
           object of :class:`genomvar.Reference` is expected.
           
           This is not necessary if self was instantiated with a reference.
           If however given, this argument takes precedence.

        Returns
        -------
        None
        """
        if isinstance(out, str):
            fh = open(out,'wt')
            opened = True
        else:
            fh = out
            opened = False

        if info_spec is None:
            if hasattr(self, 'dtype'):
                info_spec = [(v['name'], v['number'], v['type'],
                              v.get('description'), v.get('source'),
                              v.get('version')) for v in self.dtype['info'].values()]
            else:
                info_spec = None
        writer = VCFWriter(
            reference if not reference is None else self.reference,
            info_spec=info_spec)

        fh.write(writer.get_header())
        
        for vrt in self.find_vrt(expand=True):
            try:
                row = writer.get_row(vrt)
            except ValueError as exc: # TODO more specific error
                if vrt.is_instance(variant.Haplotype) \
                        or vrt.is_instance(variant.Asterisk):
                    continue
                else:
                    raise exc
            fh.write(str(row)+'\n')

        if opened:
            fh.close()

    def get_factory(self,normindel=False):
        """Returns a VariantFactory object. It will inherit the same reference
        as was used for instantiation of ``self``"""
        try:
            return self._fac['norm' if normindel else 'nonorm']
        except KeyError:
            vf = VariantFactory(self.reference,normindel=normindel)
            self._fac['norm' if normindel else 'nonorm'] = vf
            return vf

class VariantSet(VariantSetBase):
    """
    Immutable class for storing variants in-memory.

    Supports random access by variant location.
    Can be instantiated from a VCF file (see
    :meth:`VariantSet.from_vcf`) or from iterable
    containing variants (see :meth:`VariantSet.from_variants`).

    Attributes
    ----------
    reference: Reference
        Genome reference.
    ctg_len: dict
        dict mapping Chromosome name to corresponding lengths.
        
        This keeps track of the rightmost position of variants in the set
        if reference was not provided. If reference was given ``ctg_len`` 
        is taken from length of chromosomes in the reference.
    """
    def __init__(self,variants, attrib=None, vcf_notation=None, info=None,
                 sampdata=None, reference=None, dtype=None):
        super().__init__(reference=reference)
        self._variants = variants
        if not attrib is None:
            self._attrib = attrib
        if not dtype is None:
            self.dtype = dtype
        self._info = info
        self._sampdata = sampdata
        self._vcf_notation = vcf_notation
        fields = ('ind','haplotype','chrom','start','end')
        for ind,hap,chrom,start,end in zip(
                    *[self._variants[f] for f in fields]):
            if hap!=SINGLETON:
                continue
            t = self._data.setdefault(chrom, ITree())
            t.insert(start, end, value=ind)
            if not reference:
                self.ctg_len[chrom] = max(self.ctg_len.get(chrom,0),end)
                self.chroms.add(chrom)
            else:
                try:
                    if end > self.ctg_len[chrom]:
                        msg = 'Variant at {}:{}-{} is out of reference range'\
                            .format(chrom,start,end)
                        raise ValueError(msg)
                except KeyError as exc:
                    if not chrom in self.chroms:
                        msg = 'Chromosome {} not in reference'.format(chrom)
                        raise ValueError(msg)

    def _genom_vrt_from_row(self,row):
        base = self._vrt_from_row(row)
        attrib = self._get_attrib(row[0])
        return variant.GenomVariant(base=base,attrib=attrib)

    # def _vrt_from_row(self,row):

    def _get_attrib(self, rnum):
        if hasattr(self, '_attrib'):
            return self._attrib[rnum]

        EXPOSED_ATTRIB = ['id', 'qual', 'filter']
        if not self._info is None:
            attrib = {'info':self._info[rnum]}
        else:
            attrib = {}
        if not self._vcf_notation is None:
            d = dict(zip(dtype1.names,
                         self._vcf_notation[rnum].tolist()))
            d2 = {k:d.pop(k) for k in EXPOSED_ATTRIB}
            attrib['vcf_notation'] = d
            attrib.update(d2)
        if not self._sampdata is None:
            attrib.update({'samples':{s:self._sampdata[s][rnum]\
                                      for s in self._sampdata}})
        return attrib

        
    def _vrt_from_row(self,row):
        ind,hap,chrom,start,end,ref,alt,cls,start2,end2 = row
        try:
            return cls(chrom=chrom,start=start,end=end,ref=ref,alt=alt)
        except TypeError as exc:
            if cls.is_subclass(variant.AmbigIndel):
                return cls(chrom=chrom,start=(start,start2),
                           end=(end,end2),ref=ref,alt=alt)
            elif cls.is_subclass(variant.Haplotype):
                vrt = list(self._get_hap_variants(ind))
                return variant.Haplotype(chrom,vrt)
            else:
                raise exc
                
    @lru_cache(maxsize=1000)
    def _get_vrt(self,rnum):
        """Constructs a variant from row number.
        If haplotypes its subvariants will be GenomVariant instances."""
        return self._vrt_from_row(self._variants[rnum])

    @lru_cache(maxsize=1000)
    def _get_vrt(self,rnum):
        """Constructs a variant from row number.
        If haplotypes its subvariants will be GenomVariant instances."""
        ind,hap,chrom,start,end,ref,alt,cls,start2,end2 = self._variants[rnum]
        try:
            return cls(chrom=chrom,start=start,end=end,ref=ref,alt=alt)
        except TypeError as exc:
            if cls.is_subclass(variant.AmbigIndel):
                return cls(chrom=chrom,start=(start,start2),
                           end=(end,end2),ref=ref,alt=alt)
            elif cls.is_subclass(variant.Haplotype):
                vrt = list(self._get_hap_variants(ind))
                return variant.Haplotype(chrom,vrt)
            else:
                raise exc

    @lru_cache(maxsize=1000)
    def _get_genom_vrt(self,rnum):
        """Like _get_vrt but also adds attributes."""
        base = self._get_vrt(rnum)
        attrib = self._get_attrib(rnum)
        return GenomVariant(base=base,attrib=attrib)

    @staticmethod
    def _get_tups(vrt_iter):
        """Constructs tuples from variants."""
        cnt = 0
        for vrt in vrt_iter:
            if vrt.is_instance(variant.Haplotype):
                ind = cnt
                yield (ind, SINGLETON, *vrt.tolist(),
                       getattr(vrt, 'attrib', None))
                cnt += 1
                for child in vrt.variants:
                    a = child.tolist()
                    yield (cnt, ind, *a, getattr(vrt, 'attrib', None))
                    cnt += 1
            else:
                a = vrt.tolist()
                b = (cnt, SINGLETON, *a, getattr(vrt, 'attrib', None))
                yield b
                cnt += 1

    @staticmethod
    def _rows_from_variants(variants):
        """Returns rows given variants"""
        zipped = list(zip(
            *VariantSet._get_tups(variants))) # 0 for ambig
        
        _variants = np.zeros(dtype=dtype0,shape=len(zipped[0]))
        for ind,field in enumerate(dtype0.fields):
            try:
                _variants[field] = zipped[ind]
            except IndexError:
                continue
        attrib = zipped[-1]
        return _variants, attrib

    def _get_hap_variants(self,hapid):
        ind = hapid+1
        while ind<self._variants.shape[0]:
            row = self._variants[ind].tolist()
            if row[1]==hapid:
                yield self._get_genom_vrt(row[0])
            else:
                break
            ind += 1
        return

    def _find_vrt(self,chrom,start=0,end=MAX_END,expand=False):
        """Uses ITree to find variants and yields them.
        Implementation is variant set class-specific."""
        try:
            intervals = self._data[chrom].find(start,end)
        except KeyError:
            return
        for s,e,rown in intervals:
            vrt = self._get_genom_vrt(rown)
            yield vrt
            if vrt.is_instance(variant.Haplotype) and expand:
                for vrt in vrt.find_vrt(start,end):
                    yield vrt

    def _find_chrom_vrt(self,chrom,expand=False):
        try:
            ivl = self._data[chrom].iter_ivl()
        except KeyError:
            return iter([])
        inds = [i[2] for i in ivl]
        vrt = self._variants[inds]
        for row in zip(*[vrt[f] for f in self._variants.dtype.fields]):
            if row[1]!=SINGLETON:
                continue
            if row[-3]==variant.Haplotype:
                hap = self._get_genom_vrt(row[0])
                yield hap
                if expand:
                    for vrt in hap.variants:
                        yield vrt
            else:
                vrt = self._genom_vrt_from_row(row)
                yield vrt
                
    def copy(self):
        """Returns a copy of variant set"""
        cpy = self.__new__(self.__class__)
        cpy.__init__(variants=self._variants.copy(),
                     info=self._info.copy(),
                     sampdata={s:self._sampdata[s].copy() \
                               for s in self._sampdata} )
        return cpy
    def nof_unit_vrt(self):
        """Returns number of simple genomic variations in a variant set. 

        SNPs and indels count as 1. MNPs as number of affected nucleotides.
        Mixed variants (if present) are count also as 1. 
        """
        N = 0
        for start,end,cls in zip(self._variants['start'],self._variants['end'],
                                 self._variants['vartype']):
            if cls.is_subclass(variant.Haplotype):
                continue
            else:
                if cls.is_subclass(variant.Indel):
                    N += 1
                elif cls.is_subclass(variant.MNP):
                    N += (end-start)
                elif cls.is_subclass(variant.Mixed):
                    N += 1
        return N

    def drop_duplicates(self,return_dropped=False):
        """
        Remove non-unique variants.
        
        Returns a variant set where no pair of variants
        is edit-equal. Haplotypes are left as is.

        Parameters
        ----------
        return_dropped : bool
            Whether to return a list of dropped variants.  *Default: False*

        Returns
        -------
        vs : VariantSet
            Variant set with unique variants.
        dropped : list of variants, optional
            if ``return_dropped`` is True dropped variants are also returned.
        """
        ret = {}
        mask = (self._variants['haplotype']==SINGLETON)\
            &(self._variants['vartype']!=variant.Haplotype)
        dedup = self._variants[mask]
        
        vrt = {}
        fields = {}
        # Insertions
        fields['ins'] = ['chrom','start','alt']
        reg_ins = dedup[dedup['vartype']==variant.Ins].copy()
        ambig_ins = dedup[dedup['vartype']==variant.AmbigIns].copy()
        ambig_ins['start'] = ambig_ins['start2']
        vrt['ins'] = np.concatenate([reg_ins,ambig_ins])

        # Deletions
        fields['dels'] = ['chrom','start','end']
        reg_dels = dedup[dedup['vartype']==variant.Del].copy()
        ambig_dels = dedup[dedup['vartype']==variant.AmbigDel].copy()
        ambig_dels['start'] = ambig_dels['start2']
        ambig_dels['end'] = ambig_dels['end2']
        vrt['dels'] = np.concatenate([reg_dels,ambig_dels])

        # Rest
        fields['rest'] = ['vartype','chrom','start','end','alt']
        rest = dedup[(~np.isin(dedup['vartype'],[variant.AmbigIns,variant.Ins,
                                           variant.Del,variant.AmbigDel]))]\
                                           .copy()
        rest['vartype'] = [c.__name__ for c in rest['vartype']]
        vrt['rest'] = rest

        uniq = {}
        for tp in ('ins','dels','rest'):
            u,idx = np.unique(vrt[tp][fields[tp]],return_index=True)
            uniq[tp] = np.take(vrt[tp],idx)['ind']
            
        ind = np.concatenate((self._variants[~mask]['ind'],
                              *uniq.values()))

        vrt = self._variants[ind]
        vrt['ind'] = np.arange(vrt.shape[0])

        vcf_not = self._vcf_notation[ind] if not self._vcf_notation \
                     is None else None
        info = self._info[ind] if not self._info is None else None
        sampdata = {s:self._sampdata[s][ind] for s in self._sampdata} \
            if not self._sampdata is None else None
        ret = VariantSet(vrt,vcf_not,info,sampdata)
        if return_dropped:
            dropped_ind = self._variants['ind']\
                [~np.isin(self._variants['ind'],ind)]
            dropped = [self._get_genom_vrt(i) for i in dropped_ind]
            return ret,dropped
        else:
            return ret

    @classmethod
    def from_vcf(cls,vcf,reference=None,parse_info=False, parse_samples=False,
                 normindel=False,duplicates='ignore',parse_null=False):
        """
        Parse VCF variant file and return VariantSet object
        
        Parameters
        ----------
        vcf : str
            VCF file to read data from.
        reference: :class:`genomvar.Reference`, optional
            Reference genome. *Default: None*
        parse_info: bool, optional
            If ``True`` INFO fields will be parsed.
            *Default: False*.
        parse_samples: bool, str or list of str, optional
            If ``True`` all sample data will be parsed.  If string, sample 
            with this name will be parsed.  If list of strings only these
            samples will be parsed. *Default: False*
        normindel: bool, optional
            If ``True`` indels will be normalized. ``reference`` should also
            be provided.  *Default: False*
        duplicates: {'ignore','keepfirst','raise'}
            How to treat duplicate variants. If ``keepfirst`` only the first
            is taken. If ``raise`` error is raised. *Default: ``ignore`` and 
            do not check for duplicates*.
        parse_null: bool, optional
            If ``True`` null variants will also be parsed (experimental).
            *Default: False*
        Returns
        -------
        VariantSet
        """
        reader = _get_reader(vcf, reference=reference)
        records = reader.get_records(parse_info=parse_info,
                                     parse_samples=parse_samples,
                                     normindel=normindel)
        if duplicates!='ignore':
            records = cls._check_duplicates(records, duplicates=duplicates)
        # else:
        #     vrt = records['vrt']
        #     vcf_notation = records['vcf']
        #     info = records.get('info')
        #     sampdata = records.get('sampdata')
            
        vset = cls.__new__(cls)
        vset.__init__(records['vrt'],
                      vcf_notation=records['vcf'],
                      info=records.get('info'),
                      sampdata=records.get('sampdata'),
                      reference=reference,
                      dtype=reader.dataparser.dtype)
        return vset

    @classmethod
    def from_variants(cls,variants,reference=None,duplicates='ignore'):
        """
        Instantiate VariantSet from iterable of variants.
        
        Parameters
        ----------
        variants : iterable
            Variants to add.
        reference : Reference, optional
            Genome reference *Default: None*
        Returns
        -------
        VariantSet
        """
        _variants, attrib = cls._rows_from_variants(variants)
        # print('!!', _variants, attrib)
        vset = cls.__new__(cls)
        vset.__init__(_variants, attrib=attrib, reference=reference)
        return vset

    def to_records(self, nested=True):
        """
        Export ``self`` to NumPy ndarray.
        
        Parameters
        ----------

        nested : bool, optional
            Matters only if ``self`` was instatiated with INFO or SAMPDATA.
            When ``True`` dtype of the return will be nested. If ``False``
            no recursive fields will be present and "info_" or "SAMP_%SAMP%" will
            be prepended to corresponding fields *Default: True*

        Returns
        -------
        NumPy ndarray with structured dtype
        """
        dtype = [(f,self._variants.dtype.fields[f][0]) for f in \
                      ('chrom','start','end','ref','alt') ]
        dtype.append( ('vartype','U10') )
        dtype.append( ('phase_group',np.int_) )

        if hasattr(self, '_attrib'):
            dtype.append( ('attrib', np.object_) )
        else:
            if not self._info is None:
                if nested:
                    dtype.append( ('info',self._info.dtype) )
                else:
                    for field,(_dtype,offset) in self._info.dtype.fields.items():
                        dtype.append( ('info_'+field,_dtype) )
            if not self._sampdata is None:
                if nested:
                    samp_dtype = [(s,self._sampdata[s].dtype) for s in self._sampdata]
                    dtype.append( ('SAMPLES',samp_dtype) )
                else:
                    for samp in self._sampdata:
                        for field,(_dtype,offset) in \
                                    self._sampdata[samp].dtype.fields.items():
                            dtype.append( ('SAMPLES_{}_{}'.format(samp,field),
                                           _dtype) )

        ret = np.zeros(shape=self._variants.shape,dtype=dtype)

        for field in ('chrom','start','end','ref','alt','vartype'):
            ret[field] = self._variants[field]
        ret['phase_group'] = self._variants['haplotype']
        ret['vartype'] = [v.__name__ for v in self._variants['vartype']]

        mask = (ret['vartype']==variant.AmbigIns)\
            |(ret['vartype']==variant.AmbigDel)
        ret[mask]['start'] = self._variants[mask]['start2']
        ret[mask]['end'] = self._variants[mask]['end2']

        # now actually populating the data
        if hasattr(self, '_attrib'):
            ret['attrib'] = self._attrib
        else:
            if not self._info is None:
                if nested:
                    ret['info'] = self._info
                else:
                    for field in self._info.dtype.fields:
                        ret['info_'+field] = self._info[field]

            if not self._sampdata is None:
                if nested:
                    for samp in self._sampdata:
                        ret['SAMPLES'][samp] = self._sampdata[samp]
                else:
                    for samp in self._sampdata:
                        for field in self._sampdata[samp].dtype.fields:
                            ret['SAMPLES_{}_{}'.format(samp,field)] = \
                                        self._sampdata[samp][field]

        nohp_mask = self._variants['vartype']!=variant.Haplotype
        ret = ret[nohp_mask]

        ret['ref'][ret['ref']==''] = '-'
        ret['alt'][ret['alt']==''] = '-'
        return ret


    @property
    def chroms(self):
        """Chromosome set"""
        try:
            return self._chroms
        except AttributeError:
            chroms,ind = np.unique(self._variants['chrom'],return_index=True)
            self._chroms = ChromSet(self._variants['chrom'][np.sort(ind)])
            return self._chroms
    
    def sort_chroms(self,key=None):
        """
        Sorts chromosomes in the set
        
        This can influence order in which methods return variants and 
        variant set comparison.

        Parameters
        ----------
        key : callable, optional
            key to use for chromosome sorting. If not given will sort
            lexicographically. *Default: None*

        Returns
        -------
        None
        """
        if not key:
            ind = np.argsort(self._variants['chrom'])
        else:
            vals = [key(c) for c in self._variants['chrom']]
            ind = np.argsort(vals)
        
        self._variants = np.take(self._variants,ind)
        if self._info:
            self._info = np.take(self._info,ind)
        if self._sampdata:
            for s,val in self._sampdata.items():
                self._sampdata[s] = np.take(self._sampdata[s],ind)
        chroms,ind = np.unique(self._variants['chrom'],return_index=True)
        self._chroms = ChromSet(self._variants['chrom'][np.sort(ind)])

    @staticmethod
    def _iterate_aligned(vs1,vs2):
        def _iterate(ar1,ar2):
            vars1 = zip(*[ar1[f] for f in vs1._variants.dtype.fields])
            vars2 = zip(*[ar2[f] for f in vs2._variants.dtype.fields])
            while True:
                try:
                    yield next(vars1),next(vars2)
                except StopIteration:
                    return
        ind1 = []
        ind2 = []
        fields = ('ind','haplotype','chrom','start','end')
        nomatch = []
        vrt = vs1._variants[np.argsort(vs1._variants['chrom'])]
        for ind,hap,chrom,start,end in zip(
                *[vrt[k] for k in fields]):
            if hap!=SINGLETON:
                continue
            if chrom not in vs2.chroms:
                nomatch.append(ind)
                continue
            ovlp = vs2._data[chrom].find(start,end)
            if not ovlp:
                nomatch.append(ind)
            else:
                ind1.extend([ind]*len(ovlp))
                ind2.extend([v[2] for v in ovlp])

        if ind2:
            ar1 = np.take(vs1._variants,ind1)
            ar2 = np.take(vs2._variants,ind2)
            it = _iterate(ar1, ar2)
        else:
            it = iter([])
        return nomatch,it

    def _cmp(self,other,action,match_ambig=False,match_partial=True):
        """Returns a new variant set depending on operation
        specified.
        """
        def _quickly_match(v1,v2):
            """Return True if match, False if not and None if unclear
            without object instantiation"""
            ind1,h1,chrom1,start1,end1,ref1,alt1,vt1,*rest = v1
            ind2,h2,chrom2,start2,end2,ref2,alt2,vt2,*rest = v2
            if vt1.is_subclass(variant.Haplotype) \
                          or vt2.is_subclass(variant.Haplotype):
                return

            if vt1.is_subclass(variant.SNP):
                if vt2==vt1:
                    if start1==start2 and alt1==alt2:
                        return True
                    else:
                        return False
                elif vt2==variant.MNP:
                    return
                else: # Indel or Mixed
                    return False
            elif vt1==variant.Del:
                if vt2==vt1:
                    if start1==start2 and end1==end2:
                        return True
                    else:
                        return False
                elif vt2.is_subclass(variant.AmbigDel):
                    return
                else:
                    return False
            elif vt1==variant.Ins:
                if vt2==vt1:
                    if start1==start2 and alt1==alt2:
                        return True
                    else:
                        return False
                elif vt2.is_subclass(variant.AmbigIns):
                    return
                else:
                    return False
            else:
                return
            # _quicky_match ends here

        if not isinstance(other,VariantSet):
            msg = '{} should be instance of VariantSet'.format(other)
            raise TypeError(msg)

        nomatch,aligned = self._iterate_aligned(self,other)
        if action=='diff':
            ind2take = nomatch
        else:
            ind2take = []
        vrt2add = []
        for ind1,subit in itertools.groupby(aligned,key=lambda p: p[0][0]):
            vars1,locus = list(zip(*subit))
            var1 = vars1[0]
            if len(locus)==1:
                maybe_equal = _quickly_match(var1,locus[0])
            else:
                maybe_equal = None

            if not maybe_equal is None:
                if maybe_equal:
                    if action=='comm':
                        ind2take.append(ind1)
                elif action=='diff':
                    ind2take.append(ind1)
            else: # last resort, instantiating the objects
                vrt1 = self._get_vrt(var1[0])
                locus_vrt = [other._get_vrt(r[0]) for r in locus]
                mt = ops.matchv(
                    vrt1,locus_vrt,match_partial=match_partial,
                    match_ambig=match_ambig)
                comp = ops.cmpv(vrt1,mt,action)
                # TODO preserve haplotypes
                vrt2add.extend([(var1[0],v) for v in comp])

        new = np.take(self._variants,ind2take)
        if vrt2add:
            append, append_attrib = VariantSet._rows_from_variants([t[1] for t in vrt2add])
            try:
                new = np.concatenate([new,append])
            except TypeError as exc:
                if new.shape[0]==0:
                    new = append
                else:
                    raise exc
        new['ind'] = np.arange(new.shape[0])
        if not self._info is None:
            info = np.take(self._info,ind2take)
            if vrt2add:
                extra = np.take(self._info,[t[0] for t in vrt2add])
                info = np.concatenate([info,extra])
        else:
            info = None
        if not self._sampdata is None:
            sampdata = {}
            for samp in self._sampdata:
                _sampdata = np.take(self._sampdata[samp],ind2take)
                if vrt2add:
                    extra = np.take(self._sampdata[samp],
                                    [t[0] for t in vrt2add])
                    _sampdata = np.concatenate([_sampdata,extra])
                sampdata[samp] = _sampdata
        else:
            sampdata = None
        rt = self.__new__(self.__class__)
        rt.__init__(new,info=info,sampdata=sampdata)
        return rt
    
    def diff(self,other,match_partial=True,
             match_ambig=False):
        """
        Returns a new variant set object which has variants
        present in ``self`` but absent in the ``other``.

        Parameters
        ----------

        match_partial: bool
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*
        match_ambig: bool
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*

        Returns
        -------
        diff : VariantSet
            different variants

        Notes
        -----
        Here is an example of ``diff`` operation on two variant sets s1 and s2.
        Positions identical to REF are empty::

          REF         GATTGGTAC
          ---------------------
          s1          C  CCC  T
                         CCC  G
          ---------------------
          s2                  T
                        AC    T
          =====================
          s1.diff(s2) C   CC
                          CC  G
          ---------------------
          s2.diff(s1)   A

        """
        return self._cmp(other,'diff',match_partial=match_partial,
                         match_ambig=match_ambig)

    def comm(self,other,match_partial=True,match_ambig=False):
        """
        Returns a new variant set object which has variants
        present both ``self`` and ``other``.

        Parameters
        ----------
        match_partial: bool
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*
        match_ambig: bool
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*

        Returns
        -------
        comm : VariantSet
            Common variants.

        Notes
        -----
        Here is an example of ``diff`` operation on two variant sets s1 and s2.
        Positions identical to REF are empty::

          REF         GATTGGTAC
          ---------------------
          s1          C  CCC  T
                         CCC  G
          ---------------------
          s2                  T
                        AC    T
          =====================
          s1.comm(s2)    C    T
                         C
        """
        return self._cmp(other,'comm',
                         match_partial=match_partial,
                         match_ambig=match_ambig)

class MutableVariantSet(VariantSetBase):
    """
    Mutable class storing variants supporting addition of variants with
    specified genotype.
    
    Supports random access by variant location.
    Can be instantiated from a VCF file (see
    :meth:`genomvar.varset.VariantSet.from_vcf`) or from iterable containing
    variants (see :meth:`VariantSet.from_variants`).

    Attributes
    ----------
    reference: Reference
        Genome reference.
    chroms: ChromSet
        list-like representation of chromosome in the set.
    ctg_len: dict
        dict mapping Chromosome name to corresponding lengths.
        
        This keeps track of the rightmost position of variants in the set
        if reference was not provided. If reference was given ``ctg_len`` 
        is taken from length of chromosomes in reference.
    """
    def __init__(self, data=None, reference=None, max_ploidy=2):
        super().__init__(variants=None,data=data,reference=reference)
        self._obs = {}
        self.max_ploidy = max_ploidy

    def copy(self):
        """Returns a copy of variant set"""
        cpy = self.__new__(self.__class__)
        data = {c:t.copy() for c,t in self._data.items()}
        cpy.__init__(data,self.reference)
        return cpy

    @classmethod
    def from_vcf(cls,vcf,reference=None,normindel=False,
                 parse_info=False,sample=None,
                 max_ploidy=2,parse_null=False):
        """
        Parse VCF variant file and return VariantSet object
        
        Parameters
        ----------
        vcf : str
            VCF file to read data from.
        parse_info: bool, optional
            If ``True`` INFO fields will be parsed.
            *Default: False*.
        parse_samples: bool, str or list of str, optional
            If ``True`` all sample data will be parsed.  If string, sample 
            with this name will be parsed.  If list of string only these
            samples will be parsed. *Default: False*
        normindel: bool, optional
            If ``True`` indels will be normalized. ``reference`` should also
            be provided.  *Default: False*
        reference: genomvar.Reference, optional
            Reference genome. *Default: None*
        max_ploidy: int, optional
            Indicates how many variants can occupy a single locus 
            (experimental). Default: 2
        parse_null: bool, optional
            If ``True`` null variants will also be parsed (experimental).
            *Default: False*
        Returns
        -------
        VariantSet
        """
        reader = _get_reader(vcf)
        vset = cls.__new__(cls)
        vset.__init__(reference=reference,max_ploidy=max_ploidy)

        if reference is None and normindel:
                raise ValueError('Can\'t norm indels without reference')

        try:
            for vrt in reader.iter_vrt(
                    parse_samples=sample if sample else False,
                    parse_info=parse_info):
                vset.add_vrt(vrt, GT=vrt.attrib['samples'][sample]['GT'] \
                             if sample else (1,0),
                             allow_adjust_genotype=True,attrib=vrt.attrib)
        finally:
            reader.close()
        return vset

    # def _add_iv(self,chrom,start,end):
    #     """Insert new interval with ID into ``_data``"""

    def _process_gt(self,GT,ovlp,allow_adjust_genotype=False):
        """Function takes a genotype and a list of overlapping
        variants and returns a resulting genotype depending on genotype
        parameter.
        """

        # other_gt is an array of
        # genotypes from overlap
        other_gt = np.zeros(shape=(len(ovlp),
                    max(len(GT),*[len(o.GT) for o in ovlp])))
        for ind,o in enumerate(ovlp):
            other_gt[ind,0:len(o.GT)] = o.GT

        s = np.sum(other_gt,axis=0)
        # If the positions where GT has 1 others have 1 then
        # genotypes will have to be adjust.
        if s[np.nonzero(GT)[0]].sum()>0: # have to change GT
            # zr holds the positions where genotypes from overlap
            # have zeros
            if not allow_adjust_genotype:
                msg = 'GT {:s} not compatible with variants {:s}'\
                    .format(str(GT),str(ovlp))
                raise OverlappingHaplotypeVars(msg)
            zr = np.nonzero(s==0)[0]
            if len(zr)>0: # There is an empty spot to squeeze in new genotype
                GT = np.zeros(other_gt.shape[1],dtype=np.int_)
                GT[zr[0]] = 1
                GT = tuple(GT)
            else: # No space to squeze in GT
                if max([len(v.GT) for v in ovlp])+1>self.max_ploidy:
                    msg = "Can't increase ploidy more than maximum of "\
                        +"{} when inserting around {}"\
                        .format(self.max_ploidy,ovlp)
                    raise OverlappingHaplotypeVars(msg)
                for cur_vrt in ovlp:
                    cur_vrt.GT = (0,)+cur_vrt.GT
                GT = (1,)+(0,)*other_gt.shape[1]
        # Genotypes are compatible as is
        elif s[np.nonzero(GT)[0]].sum()==0:
            pass
        return GT


    def _find_vrt(self,chrom,start=0,end=MAX_END,expand=False):
        """Auxillary function invoked by ``find_vrt``"""
        for tup in self._find_iv(chrom,start,end):
            vrt = tup[-1] #self._variants[chrom][ivl_id]
            yield vrt
            if vrt.is_instance(variant.Haplotype) and expand:
                for _vrt in vrt.find_vrt(start,end):
                    yield _vrt

    def add_hap_variants(self,variants,GT,attrib=None,
                         allow_adjust_genotype=False):
        """Add variants as a single haplotype defined by genotype GT
        """
        hap = Haplotype.from_variants(variants)
        if attrib:
            for v,attr in zip(hap.variants,attrib):
                if attr:
                    v.attrib = attr
        vrt = self.add_vrt(
            hap,GT=GT,allow_adjust_genotype=allow_adjust_genotype)
        return vrt
        
    def nof_unit_vrt(self,collapse_homozigosity=True):
        """
        Returns number of simple genomic variations in a variant set. 
        
        SNPs and indels count as 1. MNPs as number of affected nucleotides.

        Parameters
        ----------
        collapse_homozigosity: bool
            If ``True`` same unit variants on different chromosomes 
            are counted as 1. Otherwise counted separately. *Default: True*
        Returns
        -------
        int
        """
        N = 0
        variants = self.find_vrt(expand=True)
        if collapse_homozigosity:
            it1,it2,it3 = itertools.tee(variants,3)
            mnps = filter(lambda v: v.is_instance(variant.MNP),it1)
            insertions = filter(lambda v: v.is_instance(variant.Ins),it2)
            deletions = filter(lambda v: v.is_instance(variant.Del),it3)
            # for grp in no_ovlp(mnps):
            #     if len(grp)==1:
            #         N += grp[0].nof_unit_vrt()
            #     else:
            #         N += ops.nof_snp_vrt(grp)

            for _,it in itertools.groupby(insertions,lambda v: (v.start,v.alt)):
                N += 1
            for _,it in itertools.groupby(deletions,lambda v: (v.start,v.ref)):
                N += 1
            for grp in no_ovlp(mnps):
                if len(grp)==1:
                    N += grp[0].nof_unit_vrt()
                else:
                    N += ops.nof_snp_vrt(grp)
        else:
            for vrt in filter(lambda v: not v.is_instance(variant.Haplotype),
                              variants):
                N += vrt.nof_unit_vrt()
        return N

    def add_vrt(self,vrt,GT,allow_adjust_genotype=False,
                attrib=None):
        """
        Insert a variant into variant set.
        
        Parameters
        ----------
        vrt : VariantBase-like
            variant to insert
        GT : tuple
            Genotype of the variant, e.g. (0,1)
        allow_adjust_genotype : bool
            If the variant will overlap by genotype with variants already
            present in the set, setting to ``True`` will allow new 
            variant to fit.
        attrib : dict, optional
            Dictionary with attributes to attached to return
        Returns
        -------
        variant.GenomVariant
        """
        ovlp = self.ovlp(vrt,match_ambig=False)

        if ovlp:
            GT2 = self._process_gt(GT,ovlp,
                                   allow_adjust_genotype=allow_adjust_genotype)
        else:
            GT2 = GT

        # Initializing GenomVariant object with the VariantBase
        ret = GenomVariant(base=vrt if isinstance(vrt,VariantBase) \
                           else vrt.base,GT=GT2,attrib=attrib)

        # Add variant to interval tree and variant dict
        t = self._data.setdefault(ret.chrom, ITree())
        t.insert(ret.start, ret.end, value=ret)
        self.ctg_len[ret.chrom] = max(
            ret.end,self.ctg_len.get(ret.chrom,0))
        self._chroms.add(ret.chrom)
        return ret

    def sort_chroms(self,key=None):
        self._chroms = ChromSet(sorted(self._chroms,key=key))

    @property
    def chroms(self):
        """Chromosome set"""
        return self._chroms

    def _cmp(self,other,action,match_partial=True,match_ambig=False):
        """Returns a new variation set depending on operation
        specified.
        """
        rt = self.__new__(self.__class__)
        rt.__init__(reference=self.reference)

        for chrom in self.chroms:
            for vrt_self in self.find_vrt(chrom):
                # TODO handle null vrt
                if vrt_self.is_instance(variant.Null):
                    continue
                match = other.match(vrt_self,match_partial=match_partial,
                                    match_ambig=match_ambig)
                vrt2add = ops.cmpv(vrt_self,match,action=action)
                if not vrt2add:
                    continue
                if vrt_self.is_instance(variant.Haplotype):
                    rt.add_hap_variants(vrt2add,GT=vrt_self.GT,
                                        allow_adjust_genotype=True,
                                        attrib = [vrt_self.attrib]*len(vrt2add))
                else:
                    for _vrt in vrt2add:
                        nop = rt.add_vrt(_vrt, GT=vrt_self.GT,
                                         allow_adjust_genotype=True,
                                         attrib=vrt_self.attrib)
        return rt

    def diff(self,other,match_partial=True,
             match_ambig=False):
        """
        Returns a new variant set object which has variants
        present in ``self`` but absent in the ``other``.

        Parameters
        ----------

        match_partial: bool
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*
        match_ambig: bool
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*

        Returns
        -------
        diff : MutableVariantSet
            Different variants.

        Notes
        -----
        Here is an example of ``diff`` operation on two variant sets s1 and s2.
        Positions identical to REF are empty::

          REF         GATTGGTAC
          ---------------------
          s1          C  CCC  T
                         CCC  G
          ---------------------
          s2                  T
                        AC    T
          =====================
          s1.diff(s2) C   CC
                          CC  G
          ---------------------
          s2.diff(s1)   A

        """
        return self._cmp(other,'diff',match_partial=match_partial,
                         match_ambig=match_ambig)

    def comm(self,other,match_partial=True,match_ambig=False):
        """
        Returns a new variant set object which has variants
        present both ``self`` and ``other``.

        Parameters
        ----------
        match_partial: bool
            If ``False`` common positions won't be searched in non-identical
            MNPs. *Default: True*
        match_ambig: bool
            If ``False`` ambigous indels will be treated as regular indels.
            *Default: False*

        Returns
        -------
        comm : MutableVariantSet
            Common variants

        Notes
        -----
        Here is an example of ``diff`` operation on two variant sets s1 and s2.
        Positions identical to REF are empty::

          REF         GATTGGTAC
          ---------------------
          s1          C  CCC  T
                         CCC  G
          ---------------------
          s2                  T
                        AC    T
          =====================
          s1.comm(s2)    C    T
                         C
        """
        return self._cmp(other,'comm',
                         match_partial=match_partial,
                         match_ambig=match_ambig)

class CmpSet(object):
    """
    Class ``CmpSet`` is a lazy implementation of variant set comparison which is
    returned by methods ``diff_vrt`` and ``comm_vrt`` of any variant set class
    instance. Useful for comparisons including large VCF files while keeping
    memory profile low.
    """
    def __init__(self,left,right,op,match_partial=True,match_ambig=False):
        self.left = left
        self.right = right
        self.action = op
        self.match_partial = match_partial
        if (isinstance(self.left,VariantFileSet) \
                 or isinstance(self.right,VariantFileSet)) and match_ambig:
            raise ValueError('Cant match ambigous indels on VariantFileSet'\
                             +' subclasses')
        self.match_ambig = match_ambig

    def _ordered_chroms(self,check_consistency=True):
        """Returns a list indicating how chromosomes should be iterated
        to compare the sets.

        Returns
        -------
        (which, chrom) : (int, str)
           if ``which`` is 0 left should be advanced, if 1 then right,
           if 2 then both. ``chrom`` is the chromosome.
        """
        def _split_at(it,pred):
            buf = []
            for item in it:
                if pred(item):
                    yield buf
                    buf = []
                else:
                    buf.append(item)
            yield buf

        lchrom = self.left.get_chroms()
        rchrom = self.right.get_chroms()
        
        comm_chrom = list(filter(lambda c: c in rchrom, lchrom))
        if check_consistency:
            comm_chrom2 = list(filter(lambda c: c in lchrom, rchrom))
            if comm_chrom != comm_chrom2:
                raise DifferentlySortedChromsError
        

        chroms = []
        
        pred = lambda e: e in comm_chrom
        for left,right,common in zip(_split_at(lchrom,pred),
                                     _split_at(rchrom,pred),
                                     comm_chrom+[None]):
            chroms.extend([(0,e) for e in left]\
                          +[(1,e) for e in right]\
                          +([(2,common)] if not common is None else []))

        return chroms
    
    def region(self,chrom=None,start=0,end=MAX_END,rgn=None):
        """
        Generates variants corresponding to comparison within a region.
        
        Region can be specified using chrom,start,end or rgn with notation like
        ``chr15:1000-2000``

        Parameters
        ----------
        chrom : str
            Chromosome
        start : int, optional
            Start of region. *Default: 0*
        end: int
            End of region. *Default: max chrom length*
        rgn : str
            Region to yield variants from

        Yields
        ------
        :class:`genomvar.variant.GenomVariant`
        """
        if not rgn is None:
            if chrom:
                raise ValueError('rgn or chrom,start,end should be specified')
            chrom,start,end = rgn_from(rgn)
        kwds = {'chrom':chrom,'start':start,'end':end,
                  'check_order':True,}
        for vs in (self.left,self.right):
            if not isinstance(vs,IndexedVariantFileSet) \
                              and not isinstance(vs,MutableVariantSet) \
                              and not isinstance(vs,VariantSet):
                raise ValueError('{} does not support region access'\
                                 .format(repr(vs)))
        return self._iter_chrom_vrt(self.left.find_vrt(**kwds),
                                    self.right.find_vrt(**kwds))

    def iter_vrt(self,callback=None):
        """
        Generates variants corresponding to comparison.

        Parameters
        ----------
        callback: callable
            A function to be called on a match in right variants set 
            for every variant of left variant set. Result will be stored 
            in ``vrt.attrib['cmp']`` of yielded variant

        Yields
        ------
        variants : :class:`genomvar.variant.GenomVariant` or \
            tuple (int, :class:`genomvar.variant.GenomVariant`)
            if ``action`` is ``diff`` or ``comm`` function yields instances of 
            :class:`genomvar.variant.GenomVariant`. If ``action`` is ``all``
            function yields tuples where ``int`` is 0 if variant is present 
            in first set; 1 if in second; 2 if in both.
        """
        lchrom_it = self.left.iter_vrt_by_chrom(check_order=True)
        rchrom_it = self.right.iter_vrt_by_chrom(check_order=True)
        _vs = {0:[self.left], 1:[self.right], 2:[self.left,self.right]}
        try:
            if self.action=='diff':
                for which,chrom in self._ordered_chroms():
                    if which==0:
                        lchrom,it = next(lchrom_it)
                        for b in it:
                            yield b
                    elif which==1:
                        next(rchrom_it)
                    elif which==2:
                        lchrom,lit = next(lchrom_it)
                        rchrom,rit = next(rchrom_it)
                        for b in self._iter_chrom_vrt(
                                lit,rit,callback=callback):
                            yield b
            elif self.action=='comm':
                for which,_ in self._ordered_chroms():
                    if which==0:
                        next(lchrom_it)
                    elif which==1:
                        next(rchrom_it)
                    elif which==2:
                        lchrom,lit = next(lchrom_it)
                        rchrom,rit = next(rchrom_it)
                        for b in self._iter_chrom_vrt(
                                lit,rit,callback=callback):
                            yield b
            elif self.action=='all':
                for which,_ in self._ordered_chroms():
                    if which==0:
                        lchrom,it = next(lchrom_it)
                        for b in it:
                            yield (0,b)
                    elif which==1:
                        rchrom,it = next(rchrom_it)
                        if self.action=='all':
                            for b in it:
                                yield (1,b)
                    elif which==2:
                        lchrom,lit = next(lchrom_it)
                        rchrom,rit = next(rchrom_it)
                        for b in self._iter_chrom_vrt(
                                lit,rit,callback=callback):
                            yield b
        finally:
            lchrom_it.close()
            rchrom_it.close()
                
    def _iter_chrom_vrt(self,lit,rit,callback=None):
        """Subroutine of :meth:`CmpSet.iter_vrt` acting on sorted variants from
        the same chromosome."""
        def _compare(vrt,ovlp,action):
            return ops.cmpv(vrt,mt,action,callback=callback)
        
        cur_end = {}
        cur_right = []
        cur_pos_right = None
        for ind,vrt,ovlp in zip_variants(lit,rit):
            if self.action=='diff':
                if ind!=0:
                    continue
                if not ovlp:
                    yield vrt
                else:
                    mt = ops.matchv(vrt,ovlp)
                    for vrt in _compare(vrt,mt,self.action):
                        yield vrt
            elif self.action=='comm':
                if ind!=0 or not ovlp:
                    continue
                else:
                    mt = ops.matchv(vrt,ovlp)
                    for vrt in _compare(vrt,ovlp,self.action):
                        yield vrt
            elif self.action=='all':
                if not ovlp:
                    yield(ind,vrt)
                else:
                    mt = ops.matchv(vrt,ovlp)
                    for op in _compare(vrt,mt,'diff'):
                        yield (ind,op)
                    if ind!=0:
                        continue
                    for op in _compare(vrt,mt,'comm'):
                        yield (2,op)
                
class VariantFileSet(VariantSetBase):
    """
    Variant set representing variants contained in underlying VCF file specified
    on initialization.

    Attributes
    ----------
    file : str
        Variant file storing variants.
    """
    # Setting methods not supposed to be implemented to None
    # because official Python docs say so
    comm = None
    diff = None
    _find_vrt = None

    def __init__(self,file,reference=None,parse_info=False,
                 parse_samples=False):
        self._reader=_get_reader(file)
        self.parse_info = parse_info
        self.parse_samples = parse_samples
        super().__init__(reference=reference)

    def get_factory(self):
        """Returns a VariantFactory object. It will inherit the same reference
        as was used for instantiation of ``self``"""
        return super().get_factory(normindel=False)

    def iter_vrt_by_chrom(self,check_order=False):
        """Iterates variants grouped by chromosome."""
        return self._reader.iter_vrt_by_chrom(
                parse_info=self.parse_info,
                parse_samples=self.parse_samples,
                check_order=check_order)

    def iter_vrt(self,expand=False):
        """
        Iterate over all variants in underlying VCF.
        
        Parameters
        ----------
        expand : bool, optional
            If ``True`` variants in encountered haplotypes are yielded.
            *Default: False*
        Yields
        -------
        variant : :class:`genomvar.variant.GenomVariant`
        """
        return self._reader.iter_vrt(
            parse_info=self.parse_info,
            parse_samples=self.parse_samples)

    def nof_unit_vrt(self):
        # """
        # Returns number of simple genomic variations in a variant set. 
        
        # Returns
        # -------
        # N : int

        # Notes
        # -----
        # See documentation on ref:`variant_scoring` for details.
        # """
        N = 0
        variants = self.find_vrt(expand=True)
        for vrt in filter(lambda v: not v.is_instance(variant.Haplotype),
                          variants):
            N += vrt.nof_unit_vrt()
        return N

    def get_chroms(self):
        return self._reader.get_chroms(allow_no_index=True)

class IndexedVariantFileSet(VariantFileSet):
    """
    Variant set representing variants contained in underlying VCF file specified
    on initialization. By default it searches for Tabix index and uses it 
    for random access of variants.
    """
    def __init__(self,file,index=None,parse_info=False,parse_samples=False,
                 reference=None):
        super().__init__(file,reference=reference,parse_info=parse_info,
                         parse_samples=parse_samples)
        self._reader=VCFReader(file,index=True)
        self._chroms = set(self._reader.chroms)
        
    def _find_vrt(self,chrom,start,end,expand=False):
        """Auxillary function for ``find_vrt``"""
        for vrt in self._reader.find_vrt(
                chrom,start,end,parse_info=self.parse_info,
                parse_samples=self.parse_samples):
            yield vrt

    @property
    def chroms(self):
        """Chromosome set"""
        return self._chroms



import sys
import argparse
from genomvar.cmdutils import _cmp_vcf
from genomvar import OverlappingHaplotypeVars,UnsortedVariantFileError

tasks = ['compare_vcf']
options=[{'arg':'--match_partial','action':'store_true',
          'help':'Flag; if set common positions will be searched in MNPs',
          'default':False},
         {'arg':'vcf1','metavar':'VCF1','type':str,
          'help':'First VCF file'},
         {'arg':'vcf2','metavar':'VCF2','type':str,
          'help':'Second VCF file'},
         {'arg':'--out','metavar':'OUT','type':str,
          'help':'Output VCF file (STDOUT by default)'},
]
prsr=argparse.ArgumentParser(prog='compare_vcf',
             description='Output VCF containing variants'\
                                 +' from two VCFs compared')
for opt in options:
    arg = opt.pop('arg')
    prsr.add_argument(arg,**opt)

if len(sys.argv)<2 or (sys.argv[1] not in tasks):
    usage = """Usage: python3 -m genomvar compare_vcf [options]
    > python3 -m genomvar compare_vcf -h (--help)
    """.format(*tasks)
    print(usage)
    sys.exit()

args = prsr.parse_args(sys.argv[2:])
if args.out is None:
    out = sys.stdout
    file_opened = False
else:
    out = open(args.out,'w')
    file_opened = True


try:
    cnt = _cmp_vcf(f1=args.vcf1,f2=args.vcf2,
             out=out,match_partial=args.match_partial)
except UnsortedVariantFileError as exc:
    print('Error: Unsorted Variant file detected'\
          +'\nMessage was: '+exc.args[0],
          file=sys.stderr)
print('Unit variants:\n'\
      +'     first only: {}\n'.format(cnt[0])\
      +'     second only: {}\n'.format(cnt[1])\
      +'     both: {}'.format(cnt[2]),
      file=sys.stderr)
if file_opened:
    out.close()



"""
# genomvar #
Package allows to work with genomic variations in Python.
"""
import collections
import pyfaidx
import numpy as np

class OverlappingHaplotypeVars(Exception):
    def __init__(self,message,ovlp=None):
        self.message = message
        self.ovlp = ovlp
class VCFSampleMismatch(Exception):pass
class UnsortedVariantFileError(Exception): pass
class DifferentlySortedChromsError(Exception): pass
class DuplicateVariants(Exception): pass
class NoVCFNotationError(Exception): pass
class StructuralVariantError(NotImplementedError): pass
class NoIndexFoundError(Exception): pass

class ChromSet(collections.abc.Set):
    '''Chromosome set remembers the order of chromosome addition.'''
    def __init__(self, iterable=None):
        self.elements =  collections.OrderedDict()
        if not iterable is None:
            for value in iterable:
                self.elements[value] = 0

    def __iter__(self):
        return iter(self.elements)

    def __contains__(self, value):
        return value in self.elements

    def __len__(self):
        return len(self.elements)

    def add(self,value):
        self.elements[value] = 0

    def __repr__(self):
        return 'ChromSet({})'.format(repr(list(self.elements)))

class Reference(object):
    """A wrapper around pyfaidx FastaFile class.
    It supports cached fetching of reference sequence.
    Speed up is achieved by minimizing disk IO
    when consequtive queries are close to one another.
    """
    def __init__(self, filename, cache_dst=1000000):
        self.filename = filename
        self.REF = pyfaidx.Fasta(self.filename)
        self.ctg_len = {k:len(v) for k,v in self.REF.records.items()}
        self.cache_dst = cache_dst
        self.cache = {}

    def get(self,chrom,start,end):
        """Reference object uses cache to improve performance on consequtive
        locations by reducing IO."""
        if start<0:
            raise IndexError('start should be >=0')
        if end - start >= self.cache_dst: # Large sequence is requested
            return str(self.REF[chrsom][start:end])
        c = self.cache.get(chrom,None)
        if c: # if chrom in cache, try it
            start2 = start - c[0] # c[0] is a start of seq in cache
            end2 = end - c[0]
            if start2>=0 and end2<2*self.cache_dst-1: # within cached
                ret = c[1][start2:end2]
                return ret
            else: # new location on chrom, new cache for chrom
                cstart = max(start-self.cache_dst,0)
                cend = cstart+2*self.cache_dst
                refseq = str(self.REF[chrom][cstart:cend])
                self.cache[chrom] = (cstart,refseq)
                return refseq[start-cstart:end-cstart]
        else:
            cstart = max(start-self.cache_dst,0)
            cend = cstart+2*self.cache_dst
            refseq = str(self.REF[chrom][cstart:cend])
            self.cache[chrom] = (cstart,refseq)
            return refseq[start-cstart:end-cstart]

    def close(self):
        self.REF.close()

    def get_chroms(self):
        return list(self.REF.records)

SINGLETON = -1

MAX_END = np.iinfo(np.int32).max # Looks like it is max for bx-python



"""Module contains classes needed for VCF parsing.

Main class is :class:`VCFReader` which is instantiated from VCF file. 
VCF can be gzipped. Bgzipping and tabix-derived indexing is also 
supported for random coordinate-based access.

:class:`VCFReader` class can iterate over rows which are tuple-like object
containing VCF field strings as attributes without conversion (except
POSition which is converted to int).

Alternatively iteration of variants is supported. It yields
:class:`genomvar.variant.GenomVariant` objects.

    >>> reader = VCFReader('example.vcf.gz')
    >>> vrt = next(reader.iter_vrt(parse_info=True, parse_samples=True))
    >>> print(vrt)
    <GenomVariant: Del chr24:23-24 G/->
    >>> print(vrt.attrib['info'])
    {'NSV': 2, 'AF': 0.5, 'DP4': (11, 22, 33, 44), 'ECNT': 1,\
    'pl': 3, 'mt': 'SUBSTITUTE', 'RECN': 18, 'STR': None}

If ``parse_info`` and ``parse_samples`` parameters are ``True`` Then
INFO and SAMPLEs fields contained in VCF are parsed and split
corresponding to an allele captured by variant object. For performance
reasons these parameters are set to False and these fields are not
parsed.

.bcf format is supported, use :class:`BCFReader`.

"""

import os
import warnings
import heapq
from itertools import dropwhile,groupby,repeat,takewhile,\
    zip_longest,islice,chain,repeat
from collections import namedtuple,OrderedDict,deque
import re
import gzip
import numpy as np
from genomvar import Reference,SINGLETON,StructuralVariantError,\
    variant,ChromSet,VCFSampleMismatch,\
    UnsortedVariantFileError,MAX_END,NoIndexFoundError
from genomvar.utils import rgn_from, grouper
from genomvar.variant import GenomVariant,VariantFactory
import genomvar
from genomvar.vcf_utils import (VCFRow, VCFInfoSpec, VCF_info_fields,
                                _make_field_writer_func,
                                header as vcf_header,
                                dtype2string, string2dtype,
                                field_writer_simple)


# data type of main variants array of VariantSet class
dtype0 = np.dtype([('ind',np.int_),('haplotype',np.int_),('chrom','O'),
                   ('start',np.int_),('end',np.int_),('ref','O'),('alt','O'),
                   ('vartype','O'),('start2',np.int_),('end2',np.int_)])
# data type of VCF fields array of VariantSet class
dtype1 = np.dtype([('start',np.int_),('ref',np.object_),('alt',np.object_),
                   ('id',np.object_),('qual',np.float_),('filter',np.object_),
                   ('row',np.int_)])

def _ensure_sorted(it):
    """Ensures variants are yielded in start-sorted order"""
    try:
        v = next(it)
        start,rstart,cnt,vrt = v
    except StopIteration:
        return

    hp = [(start,rstart,cnt,vrt)]
    cur_end = vrt.end
    heapq.heapify(hp)

    while True:
        try:
            start,rstart,cnt,vrt = next(it)
        except StopIteration:
            break
        if cur_end <= rstart:
            sm = heapq.heappushpop(hp, (start,rstart,cnt,vrt) )
            yield sm[-1]
        else:
            heapq.heappush(hp, (start,rstart,cnt,vrt))
        cur_end = max(vrt.end,cur_end)

    for i in range(len(hp)):
        yield heapq.heappop(hp)[-1]

def _make_converter_func(tp,num,convert=True):
    """Returns a converter function based on TYPE and NUMBER arguments."""
    if convert==False:
        conv = lambda v: v
    else:
        if tp=='U1':
            conv = np.str_
        else:
            conv = tp
    if num=='R':
        def f(v,n):
            try:
                return conv(v[n+1])
            except IndexError:
                return None
            except ValueError as exc:
                if v[n+1]=='.':
                    return None
                else:
                    raise exc
    elif num=='A':
        def f(v,n):
            try:
                return conv(v[n])
            except IndexError:
                return None
            except ValueError as exc:
                if v[n]=='.':
                    return None
                else:
                    raise exc
    elif num==1:
        if issubclass(tp,np.bool_):
            f = lambda v,n: True
        else:
            def f(v,n):
                try:
                    return conv(v)
                except ValueError as exc:
                    if v=='.':
                        return None
                    else:
                        raise exc
                    
    elif num==0:
        if issubclass(tp,np.bool_):
            f = lambda v,n: True
        else:
            f = lambda v,n: None
    elif num=='G' or num=='.':
        f = lambda v,n: tuple(map(tp,v))
    else: # num 2,3 ...
        def f(v,n):
            try:
                return tuple([conv(v[i]) for i in range(num)])
            except IndexError as exc:
                r = []
                for i in range(num):
                    if i>len(v)-1:
                        r.append(None)
                    else:
                        r.append(conv(v[i]))
                return tuple(r)
    return f

no_converter = _make_converter_func(np.object_, 1, convert=False)

def _isindexed(file):
    idx = file+'.tbi'
    if os.path.isfile(idx):
        if os.path.getmtime(idx)>=os.path.getmtime(file):
            return True
        else:
            warnings.warn('Index is outdated for {}'.format(file))
    return False

def _check_VCF_order(it):
    """Raises UnsortedVariantFileError if row iterator is not sorted."""
    prev_row = namedtuple('first','CHROM POS')(object(),0)
    for row in it:
        if row.CHROM==prev_row.CHROM and row.POS<prev_row.POS:
            raise UnsortedVariantFileError
        else:
            prev_row = row
            yield row

gt_cache = {}
def _parse_gt(gt,ind):
    """Function return genotype tuple given genotype string and allele index.
    For performance results are cached based on arguments"""
    try:
        return gt_cache[(gt,ind)]
    except KeyError:
        vals = re.split('[/|]',gt)
        GT = []
        for val in vals:
            if val == '.':
                GT.append(None)
            elif int(val)==ind+1:
                GT.append(1)
            else:
                GT.append(0)
        GT = tuple(GT)
        gt_cache[(gt,ind)] = GT
        return GT

def validate_spec(spec):
    name, 
    
class VCFWriter(object):
    """Class for writing variant to VCF format."""
    def __init__(self, reference=None, info_spec=None):
        if isinstance(reference, str):
            self.reference = Reference(reference)
        else:
            self.reference = reference
        if not info_spec is None:
            self.writers = {}
            dtype = {}
            for spec in info_spec:
                if len(spec)<3:
                    raise ValueError(
                        'INFO spec expected to have at '\
                        +'least 3 fields while {} found'\
                        .format(len(spec)))
                _spec = VCFInfoSpec(
                    *islice(
                        chain.from_iterable(
                            [spec, repeat(None)]),
                        6))
                self.writers[_spec.NAME] = _make_field_writer_func(
                    _spec.TYPE, _spec.NUMBER)
                dt = {
                    'name' : _spec.NAME,
                    'number' : _spec.NUMBER,
                    'type' : _spec.TYPE,
                    'dtype' : string2dtype[_spec.TYPE],
                    'description' : _spec.DESCRIPTION,
                    'source' : _spec.SOURCE,
                    'version' : _spec.VERSION
                }
                dtype[_spec.NAME] = dt
            self.dtype = dtype

    def get_header(self):
        if hasattr(self, 'dtype'):
            info = [{f.lower():p.get(f.lower(), '.') for f in VCF_info_fields} \
                    for p in self.dtype.values()]
        else:
            info = {}
        header = vcf_header.render(
            info=info,
            ctg_len=self.reference.ctg_len if self.reference else {})
        return header

    @staticmethod
    def _get_filter(v):
        if v is None:
            return
        elif isinstance(v, list):
            return ';'.join(v)
        else:
            return str(v)

    def get_row(self, vrt, **kwds):
        """Formats a variant to a :class:`genomvar.vcf_utils.VCFRow` instance. 

        For indels writer with reference 
        might be needeed to costruct correct REF field.

        Parameters
        ----------
        vrt : Variant instance
            variant to get row for. 

        kwds : VCF fields
            optional. If given, these and only these parameters are 
            used to populate corresponding VCF fields: ``id``, 
            ``qual``, ``filter``, ``info``. 
            These parameters are taken as is and converted to string
            before returning a VCFRow. 
            If the keyword is not given corresponding field will be 
            populated from ``attrib`` attribute if possible.

            Any other keyword arguments are ignored.

        Returns
        -------
        row : VCFRow
            :class:`genomvar.vcf_utils.VCFRow` object instance

        Notes
        -----
        >>> factory = variant.VariantFactory()
        >>> writer = VCFWriter()
        >>> v1 = factory.from_edit('chr24', 2093, 'TGG', 'CCC')
        >>> row = writer.get_row(v1)
        >>> row
        <VCFRow chr24:2094 TGG->CCC>
        >>> print(row)
        chr24	2094	.	TGG	CCC	.	.	.

        >>> print(writer.get_row(v1, id=123, info='DP=10'))
        chr24	2094	123	TGG	CCC	.	.	DP=10
        """
        if hasattr(vrt, 'attrib'):
            pos, ref, alt = vrt._get_vcf_notation(
                vcf_notation=vrt.attrib.get('vcf_notation'),
                reference=self.reference)
            dt = vrt.attrib
            if 'info' in kwds:
                info = kwds['info']
            elif 'info' in dt and not dt['info'] is None:
                info = self._format_info(dt['info'])
            else:
                info = '.'
            r = VCFRow(vrt.chrom, pos,
                       kwds.get('id', dt.get('id')),
                       ref.upper(), alt.upper(),
                       kwds.get('qual', dt.get('qual')),
                       self._get_filter(kwds.get('filter', dt.get('filter'))),
                       info)
            return r
        else:
            pos, ref, alt = vrt._get_vcf_notation(reference=self.reference)
            return VCFRow(vrt.chrom, pos, kwds.get('id', '.'),
                          ref.upper(),alt.upper(),
                          kwds.get('qual', '.'), kwds.get('filter', '.'),
                          kwds.get('info', '.'))


    def _format_info(self, info):
        if info is None:
            return '.'

        if isinstance(info, np.void):
            # special case of VariantSet read-in INFO data
            _info = [self.writers[k](k,v) for k,v in zip(info.dtype.fields, info)]
        else:
            if hasattr(self, 'writers'):
                _info = [w(k,info.get(k)) for k,w in \
                         self.writers.items()]
            else:
                _info = [field_writer_simple(k,v) for k,v in info.items()]
        return ';'.join(_info)
    
class VCFReader(object):
    """Class to read VCF files."""
    def __init__(self,vcf,index=False,reference=None):
        if isinstance(vcf,str):
            self.fl = vcf
            if self.fl.endswith('.gz') or self.fl.endswith('.bgz'):
                self.compressed = True
            else:
                self.compressed = False
            self.idx_file = None
            if isinstance(index,bool):
                if index:
                    idx = self.fl+'.tbi'
                    if os.path.isfile(idx):
                        self.idx_file = idx
                    else:
                        raise NoIndexFoundError('no index for '+str(self.fl))
            elif isinstance(index,str):
                if os.path.isfile(index):
                    self.idx_file = index
                else:
                    raise OSError('{} not found'.format(index))

            if self.compressed:
                self.openfn = gzip.open
            else:
                self.openfn = open
            self.buf = self.openfn(self.fl,'rt')
            self.opened_file = True
        else: # buffer-like object
            self.buf = vcf
            self.opened_file = False
        self._vrt_start = False
        self.reference = reference
        # Init default variant factory
        self._factory = VariantFactory(reference,normindel=False)
        self.vrt_fac = {'nonorm':self._factory}
        self._dtype = {'info':OrderedDict(),'format':OrderedDict()}
        self._samples = []
        for cnt,line in enumerate(self.buf):
            if line.startswith('##INFO'):
                dat = self._parse_dtype(line)
                self._dtype['info'][dat['name']] = dat
            elif line.startswith('##FORMAT'):
                dat = self._parse_dtype(line)
                self._dtype['format'][dat['name']] = dat
            elif line.startswith('#CHROM'):
                # this should be the last header line
                self.header_len = cnt + 1

                vals = line.strip().split('\t')
                if len(vals)>9:
                    self._samples = vals[9:]
                self._vrt_start = True
            elif not line.startswith('#'):
                break

        if self.opened_file:
            self.buf.close()
        self.sample_ind = OrderedDict()
        for ind,sample in enumerate(self._samples):
            self.sample_ind[sample] = ind

    def get_factory(self,normindel=False):
        """Returns a factory based on normindel parameter."""
        try:
            return self.vrt_fac['norm' if normindel else 'nonorm']
        except KeyError:
            vf = VariantFactory(self.reference,normindel=normindel)
            self.vrt_fac['norm' if normindel else 'nonorm'] = vf
            return vf

    @staticmethod
    def _parse_dtype(line):
        """Parses VCF header and returns correct datatypes."""
        rx = '##(?:(?:INFO)|(?:FORMAT))=\<ID=(\S+),Number=([\w.]+),'\
            +'Type=(\w+),Description="(.*)"(:?,Source="(.*)")?(?:,Version=".*")?\>'
        match = re.match(rx, line)
        NAME,NUMBER,TYPE,DESCRIPTION,SOURCE,VERSION = match.groups()

        try:
            num = int(NUMBER)
        except ValueError:
            num = NUMBER
        if isinstance(num,int):
            if num==0:
                size = 1
                tp = np.bool_
            else:
                size = num
                tp = string2dtype[TYPE]
        elif num in ('A','R'):
            size = 1
            tp = string2dtype[TYPE]
        elif num in ('.','G'):
            size = 1
            tp = np.object_
        else:
            raise ValueError('Unknown NUMBER:'+num)
        return {'name':NAME, 'type':TYPE, 'dtype':tp,
                'size':size, 'number':num,
                'description':DESCRIPTION}
        
    def _sample_indices(self,samples):
        ind = {}
        for sample in samples:
            if not sample in self._samples:
                raise VCFSampleMismatch('Sample {} not found'.format(sample))
            else:
                ind[sample] = self._samples.index(sample)
        return ind

    def find_rows(self,chrom=None,start=None,end=None,
                  rgn=None):
        """Yields rows of variant file"""
        cnt = 0
        if not rgn is None:
            chrom,start,end = rgn_from(rgn)
            for row in self._query(chrom,start,end):
                yield row
        elif not chrom is None:
            for row in self._query(chrom,start=start,end=end):
                yield row
        else:
            raise ValueError('rgn or chrom,start,end should be given')

    def iter_rows(self,check_order=False):
        """Yields rows of variant file"""
        
        buf = self.get_buf()
        buf.seek(0)
        return RowIterator(self,buf,check_order)
        
    def iter_rows_by_chrom(self,check_order=False):
        """Yields rows grouped by chromosome"""
        buf = self.get_buf()
        buf.seek(0)
        return RowByChromIterator(self,buf,check_order)

    def _parse_vrt(self,row,factory,parse_info=False,parse_samples=False,
                   parse_null=False):
        """Given the row returns variant objects with alts splitted.
        INFO, SAMPLE data are correctly parsed if needed"""
        alts = row.ALT.split(',')
        # TODO avoid list structure, generator instead
        if parse_info: # Read INFO if necessary
            info = self.dataparser.get_infod(row.INFO, alts=len(alts))
        else:
            info = repeat(None)
        if parse_samples and row.SAMPLES:
            sampdata = self.dataparser.get_sampdata(row, len(alts))
        else:
            sampdata = repeat(None)

        for ind,(_alt,_info,_sampdata) in enumerate(zip(alts,info,sampdata)):
            try:
                base  = factory.from_edit(row.CHROM,row.POS-1,row.REF,_alt)
            except StructuralVariantError:
                warnings.warn(
                    'Structural variants are not yet supported, skipping row '\
                    +str(row.rnum))
                continue
            infod = _info
            # if parse_info:
            #     infod = dict(zip(self._dtype['info'],_info))
            # else:
            #     infod = {}
            if parse_samples:
                sampd = {}
                for sn,samp in enumerate(self._samples):
                    sampd[samp] = dict(zip(self._dtype['format'],_sampdata[sn]))
            else:
                sampd = {}
            # FIXME avoid dictionaries
            attrib = {'info':infod,'samples':sampd,
                      'allele_num':ind,'filter':row.FILTER.split(';'),
                      'id':row.ID if row.ID!='.' else None,
                      'qual':row.QUAL}
            attrib['vcf_notation'] = {'start' : row.POS-1,'ref' : row.REF,
                                      'row' : row.rnum}
            _vrt = GenomVariant(base, attrib=attrib)
            yield _vrt

        if parse_null: # experimental
            if split_info:
                info2 = self._parse_info2(info1,allele_num='null')
            else:
                info2 = info1
            samples2 = {s:{} for s in parse_samples}
            for sm in parse_samples:
                 for key,val in samples1[sm].items():
                     if key=='GT':
                         _gt = [int(i) for i in re.split('[/|]',
                                                         samples1[sm]['GT'])]
                         samples2[sm]['GT'] = tuple([int(i==0) for i in _gt])
                     else:
                         samples2[sm][key] = val

            base  = variant.Null(row.CHROM,row.POS-1,row.REF,'-',
                                row.POS-1+len(row.REF))
            _vrt = GenomVariant(base,attrib={'info':info2,
                                             'samples':samples2,
                                             'filter':row.FILTER.split(';'),
                                             'id':row.ID,'allele_num':'null'})
            yield _vrt

    def get_records(self,parse_info=False,parse_samples=False,normindel=False):
        """
        Returns parsed variant data as a dict of NumPy arrays with structured
        dtype.
        """
        vfac = self.get_factory(normindel=normindel)
        chunk_sz = 1000
        fchunks = {'vrt': {
                       f:[] for f in dtype0.names
                   },
                   'info':{
                       f:[] for f in self._dtype['info']
                   },
                   'sampdata':{
                       s:{f:[] for f in self._dtype['format']} \
                              for s in self._samples},
                   'vcf':{f:[] for f in dtype1.names}
        }
        vnum = 0

        rows = self.iter_rows()
        for row_chunk in grouper(rows,chunk_sz):
            haps = []
            tups = {'vrt':[],'info':[],'vcf':[],
                    'sampdata':{s:[] for s in self._samples}}
            for row in row_chunk:
                if not row: # reached last row
                    break
                alts = row.ALT.split(',')
                if parse_info: # Read INFO if necessary
                    info = self.dataparser.get_info(row.INFO,alts=len(alts))
                else:
                    info = repeat(None)
                if parse_samples and row.SAMPLES:
                    sampdata = self.dataparser.get_sampdata(row,len(alts))
                else:
                    sampdata = repeat(None)
                for _alt,_info,_sampdata in zip(alts,info,sampdata):
                    try:
                        base = vfac._parse_edit(
                            row.CHROM, row.POS-1,row.REF,_alt)
                    except StructuralVariantError:
                        warnings.warn(
                            'Structural variants are not yet supported, '\
                            +'skipping row '+str(row.rnum))
                        continue
                    if base[-1].is_subclass(variant.AmbigIndel):
                        chrom,(start,start2),(end,end2),ref,alt,cls = base
                        tups['vrt'].append( (chrom,start,end,ref,alt,\
                                             cls,start2,end2) )
                        added = 1
                        haps.append(SINGLETON)
                    else:
                        tups['vrt'].append(base)
                        added = 1
                        haps.append(SINGLETON)

                    # Adding VCF notation related fields
                    tups['vcf'].append([row.POS-1,row.REF,row.ALT,row.ID,
                                        row.QUAL if row.QUAL!='.' else None,
                                        row.FILTER,row.rnum]*added)
                    if parse_info:
                        if added==1:
                            tups['info'].append(_info)
                        else:
                            tups.extend([_info]*added)
                    if parse_samples:
                        if added==1:
                            for sn in range(len(self._samples)):
                                tups['sampdata'][self._samples[sn]].append(_sampdata[sn])
                        else:
                            for sn,samp in enumerate(self._samples):
                                tups['sampdata'][samp]\
                                    .extend([_sampdata[sn]]*added)
                    vnum += added

            zipped = {k:list(zip_longest(*tups[k],fillvalue=0)) \
                                 for k in ('vrt','info','vcf')}
            zipped['sampdata'] = {}
            for samp in self._samples:
                zipped['sampdata'][samp] = list(zip(*tups['sampdata'][samp]))
            cur_sz = len(tups['vrt'])
            for ind,field in enumerate(dtype0.names[2:]): # skip ind and hap
                a = np.zeros(dtype=dtype0.fields.get(field)[0],
                             shape=cur_sz)
                try:
                    a[:cur_sz] = zipped['vrt'][ind]
                except IndexError as exc:
                    if field=='start2':
                        break # no ambigous indels
                    else:
                        raise exc
                fchunks['vrt'][field].append(np.resize(a,cur_sz))
            fchunks['vrt']['haplotype'].append(np.asarray(haps,dtype=np.int_))

            # Add VCF notation
            for ind,field in enumerate(dtype1.names): # skip ind and hap
                a = np.zeros(dtype=dtype1.fields.get(field)[0],
                             shape=cur_sz)
                a[:cur_sz] = zipped['vcf'][ind]
                fchunks['vcf'][field].append(np.resize(a,cur_sz))
            if parse_info:
                for ind,field in enumerate(self._dtype['info']):
                    fd = self._dtype['info'][field]
                    dtype = fd['dtype'] if fd['size']==1 \
                        else (fd['dtype'],fd['size'])
                    ar = np.zeros(dtype=dtype,shape=cur_sz)
                    try:
                        ar[:] = zipped['info'][ind]
                    except TypeError as exc:
                        if issubclass(ar.dtype.type,np.int_):
                            dtype = np.float_ if fd['size']==1 \
                                else (np.float_,fd['size'])
                            ar = np.zeros(dtype=dtype,shape=cur_sz)
                            ar[:] = zipped['info'][ind]
                        else:
                            raise exc # reraising if not integer
                    fchunks['info'][field].append(ar)

            if parse_samples:
                for samp in self._samples:
                    for ind,field in enumerate(self._dtype['format']):
                        fd = self._dtype['format'][field]
                        dtype = fd['dtype'] if fd['size']==1 \
                            else (fd['dtype'],fd['size'])
                        ar = np.zeros(dtype=dtype,shape=cur_sz)
                        try:
                            ar[:] = zipped['sampdata'][samp][ind]
                        except TypeError as exc:
                            if issubclass(ar.dtype.type,np.int_):
                                dtype = np.float_ if fd['size']==1 \
                                    else (np.float_,fd['size'])
                                ar = np.zeros(dtype=dtype,shape=cur_sz)
                                ar[:] = zipped['sampdata'][samp][ind]
                            else:
                                raise exc # reraising if not integer
                        fchunks['sampdata'][samp][field].append(ar)
                
        ret = {}
        ret['vrt'] = np.zeros(dtype=dtype0,shape=vnum)
        ret['vcf'] = np.zeros(dtype=dtype1,shape=vnum)

        for field in dtype0.names[1:]: # skipping ind
            try:
                ret['vrt'][field][:] = np.concatenate(fchunks['vrt'][field])
            except ValueError as exc:
                if field == 'start2': # No Ambigous indels
                    break
                elif len(fchunks['vrt'][field])==0: # empty VCF
                    break
                else:
                    raise exc
        for field in dtype1.names: 
            try:
                ret['vcf'][field][:] = np.concatenate(fchunks['vcf'][field])
            except ValueError as exc:
                if len(fchunks['vcf'][field])==0: # empty VCF
                    break
                else:
                    raise exc
        ret['vrt']['ind'] = np.arange(vnum)
        if parse_info:
            dt = self.dataparser.get_dtype('info')
            fields = {}
            for ind,(field,vals) in enumerate(fchunks['info'].items()):
                a = np.concatenate(vals)
                fields[field] = a
                # Need to update dtype due to possible conversion to float
                fdtype = dt[ind]
                dt[ind] = tuple([fdtype[0],a.dtype,*fdtype[2:]])
            ret['info'] = np.zeros(dtype=dt,shape=vnum)
            for field in ret['info'].dtype.names:
                ret['info'][field][:] = fields[field]
        if parse_samples:
            ret['sampdata'] = {}
            for samp in self._samples:
                dt = self.dataparser.get_dtype('format')
                fields = {}
                for ind,(field,vals) in enumerate(
                        fchunks['sampdata'][samp].items()):
                    a = np.concatenate(vals)
                    fields[field] = a
                    # Need to update dtype due to possible conversion to float
                    fdtype = dt[ind]
                    dt[ind] = tuple([fdtype[0],a.dtype,*fdtype[2:]])
                ret['sampdata'][samp] = np.zeros(dtype=dt,shape=vnum)
                for field in ret['sampdata'][samp].dtype.names:
                    ret['sampdata'][samp][field][:] = fields[field]
        return ret


    def iter_vrt(self,check_order=False,parse_info=False,
                 normindel=False,parse_samples=False):
        """
        Yields variant objects.
        
        Parameters
        ----------
        parse_info : bool
            Whether INFO fields should be parsed.  *Default: False*
        parse_samples : bool
            Whether SAMPLEs dat should be parsed.  *Default: False*
        check_order : bool
            If True will raise exception on unsorted VCF rows. *Default: False*
        normindel : bool
            If True insertions and deletions will be left normalized.
            Requires a reference on :class:`VCFReader` instantiation.

        Yields
        -------
        vrt : :class:`genomvar.variant.GenomVariant`
           Variant object 
        """

        if parse_samples==True:
            samps = 'all'
        else:
            samps = self._normalize_samples(parse_samples)
        return VrtIterator(self.iter_rows(check_order),
                           self.get_factory(normindel=normindel),
                           parse_info,samps)

    def iter_vrt_by_chrom(self,parse_info=False,
                          parse_samples=False,normindel=False,
                          check_order=False):
        """
        Generates variants grouped by chromosome
        
        Parameters
        ----------
        parse_info : bool
            Incates whether INFO fields should be parsed.  *Default: False*
        parse_samples : bool
            Incates whether SAMPLE data should be parsed.  *Default: False*
        parse_samples : bool
            If True indels will be normalized. ``VCFReader`` should have been
            instantiated with reference. *Default: False*
        check_order : bool
            If True VCF will be checked for sorting. *Default: False*
        Yields
        -------
        (chrom,it) : tuple of str and iterator
            ``it`` yields :class:`variant.Genomvariant` objects
        """
        if parse_samples==True:
            samps = 'all'
        else:
            samps = self._normalize_samples(parse_samples)
        return VrtByChromIterator(self.iter_rows(check_order),
                                  self.get_factory(normindel=normindel),
                                  parse_info,samps)

    def _normalize_samples(self,parse_samples):
        if isinstance(parse_samples,str):
            if not parse_samples in self._samples:
                raise VCFSampleMismatch('Sample {} not found'\
                                        .format(parse_samples))
            samples = [parse_samples]
        elif isinstance(parse_samples,list):
            for samp in parse_samples:
                if not samp in self._samples:
                    raise VCFSampleMismatch('Sample {} not found'\
                                            .format(samp))

            samples = parse_samples
        elif isinstance(parse_samples,bool):
            if parse_samples:
                samples = self._samples
            else:
                samples = []
        else:
            raise ValueError('parse sample should be str,list-like or bool')
        return samples

    def _variants_from_rows(self,it,factory,parse_info,parse_samples):
        cnt = 0
        for row in it:
            for vrt in self._parse_vrt(row,factory,parse_info,
                                       parse_samples):
                yield (vrt.start,row.POS-1,cnt,vrt)
                
                cnt += 1

    def find_vrt(self,chrom=None,start=0,end=MAX_END,
                 check_order=False,parse_info=False,normindel=False,
                 parse_samples=False):
        """Yields variant objects from a specified region"""
        factory = self.get_factory(normindel=normindel)
        # -1 for the case of insertions which are leftier in VCF notation
        # than they are in `genomvar` notation
        _rows = self.find_rows(chrom=chrom,
                               start=start-1 if start else None,
                               end=end)
        if check_order:
            rows = _check_VCF_order(_rows)
        else:
            rows = _rows
        if parse_samples==True:
            samps = 'all'
        else:
            samps = self._normalize_samples(parse_samples)

        _variants = self._variants_from_rows(rows,factory,parse_info,samps)
        for vrt in _ensure_sorted(_variants):
            if vrt.start>=end or vrt.end<=start:
                continue
            yield vrt

    def _get_tabix(self):
        if hasattr(self, 'tabix'):
            return self.tabix
        else:
            if self.idx_file:
                try:
                    self.tabix = pysam.TabixFile(
                        filename=self.fl, index=self.idx_file)
                except NameError:
                    import pysam
                    pysam.set_verbosity(0)
                    self.tabix = pysam.TabixFile(
                        filename=self.fl, index=self.idx_file)
                except OSError as exc:
                    msg = exc.args[0]
                    if 'index' in msg and 'not found' in msg:
                        raise NoIndexFoundError
                    else:
                        raise exc
            else:
                raise NoIndexFoundError
        return self.tabix
            
    def _query(self,chrom,start=None,end=None):
        """User tabix index to fetch VCFRow's"""
        if start and not end:
            raise ValueError('"start" was given but "end" was not')
        tabix = self._get_tabix()
        try:
            lines = tabix.fetch(chrom,start,end)
        except ValueError: # pysam raises on wrong chromosome
            return
        for line in lines:
            yield VCFRow(*line.strip().split('\t',maxsplit=9))

    def get_chroms(self,allow_no_index=False):
        """Returns ``ChromSet`` corresponding to VCF. If indexed 
        then index is used for faster access. Alternatively if ``allow_no_index``
        is True the whole file is parsed to get chromosome ordering."""
        try:
            tabix = self._get_tabix()
            self._chroms = ChromSet(tabix.contigs)
            return self._chroms
        except NoIndexFoundError:
            if allow_no_index:
                self._chroms = ChromSet()
                is_comment = lambda l: l.startswith('#')
                with self.openfn(self.fl,'rt') as fh:
                    lines = dropwhile(is_comment,fh)
                    for line in lines:
                        self._chroms.add(line.split('\t',maxsplit=1)[0])
                return self._chroms
            else:
                msg = 'Need index to get chroms'
                raise NotImplementedError(msg)

    def get_buf(self):
        if self.opened_file:
            r = self.openfn(self.fl,'rt')
            return r
        else:
            return self.buf
        
    def close(self):
        if self.opened_file:
            self.buf.close()
    
    @property
    def chroms(self):
        if hasattr(self,'_chroms'):
            return self._chroms
        else:
            self.get_chroms()
            return self._chroms

    @chroms.setter
    def chroms(self,value):
        raise NotImplementedError

    @chroms.deleter
    def chroms(self):
        delattr(self,'_chroms')

    @property
    def dataparser(self):
        try:
            return self._dataparser
        except AttributeError:
            self._dataparser = _DataParser(self._dtype,self.sample_ind)
            return self._dataparser

    @dataparser.setter
    def dataparser(self,value):
        raise NotImplementedError

    @dataparser.deleter
    def dataparser(self):
        delattr(self,'_dataparser')
        
class BCFReader(VCFReader):
    def __init__(self,bcf,index=False,reference=None):
        # TODO leverage inheritance
        self.fl = bcf
        self.idx_file = None
        if isinstance(index,bool):
            if index:
                idx = self.fl+'.csi'
                if os.path.isfile(idx):
                    self.idx_file = idx
                else:
                    raise OSError('Index not found')
        elif isinstance(index,str):
            if os.path.isfile(index):
                self.idx_file = index
            else:
                raise OSError('{} not found'.format(index))

        self.reference = reference
        # Init default variant factory
        self._factory = VariantFactory(reference,normindel=False)
        self.vrt_fac = {'nonorm':self._factory}
        self._dtype = {'info':OrderedDict(),'format':OrderedDict()}
        buf = self.get_buf()
        for cnt,line in enumerate(str(buf.header).splitlines()):
            if line.startswith('##INFO'):
                dat = self._parse_dtype(line)
                self._dtype['info'][dat['name']] = dat
            elif line.startswith('##FORMAT'):
                dat = self._parse_dtype(line)
                self._dtype['format'][dat['name']] = dat
            elif line.startswith('#CHROM'):
                # this should be the last header line
                self.header_len = cnt + 1

                vals = line.strip().split('\t')
                if len(vals)>9:
                    self._samples = vals[9:]
                else:
                    self._samples = []
                self._vrt_start = True
                break

        self.sample_ind = OrderedDict()
        for ind,sample in enumerate(self._samples):
            self.sample_ind[sample] = ind

    def iter_rows(self, check_order=None):
        return BCFRowIterator(self, self.get_buf(),
                              check_order=check_order)

    def get_buf(self):
        try:
            f = pysam.VariantFile(self.fl)
        except NameError:
            import pysam
            pysam.set_verbosity(0)
            f = pysam.VariantFile(self.fl)
        return f

    @property
    def dataparser(self):
        try:
            return self._dataparser
        except AttributeError:
            self._dataparser = _BCFDataParser(self._dtype,self.sample_ind)
            return self._dataparser

    @dataparser.setter
    def dataparser(self,value):
        raise NotImplementedError

    @dataparser.deleter
    def dataparser(self):
        delattr(self,'_dataparser')

class _DataParser(object):
    """Object for parsing INFO and SAMPLES data."""
    def __init__(self,dtype,sample_ind):
        self.dtype = dtype
        self.converters = {'info':{},'format':{}}
        self.none = {'info':[],'format':[]}
        for hh in ['info','format']:
            for field,props in self.dtype[hh].items():
                tp = props['dtype']
                sz = props['size']
                num = props['number']
                self.converters[hh][field] = _make_converter_func(tp,num)
                self.none[hh].append(None if sz<=1 else [None]*sz)
        self.converters['format']['GT'] = \
            lambda v,a: _parse_gt(v,a)
        self.order = {
            'info':{f:ind for ind,f in enumerate(self.dtype['info'])},
            'format':{f:ind for ind,f in enumerate(self.dtype['format'])}
        }
        self.sample_ind = sample_ind

    def tokenize_info(self,INFO):
        """Function splits info on simple key value pairs"""
        info = []
        for field in INFO.split(';'):
            if not field or field=='.':
                continue
            keyval = field.split('=',maxsplit=1)
            if len(keyval)==1:
                info.append((field, None))
            else:
                key,val=keyval
                try:
                    num = self.dtype['info'][key]['number']
                except KeyError:
                    num=1
                if num in [0,1]:
                    info.append( (key, val) )
                else:
                    info.append( (key, val.split(',')) )
        return info
        
    def get_info(self,INFO,alts=1,parse_null=False):
        """Given INFO string and number of alt alleles returns a list
        of lists with data corresponding to alleles, then fields."""
        tokenized = self.tokenize_info(INFO)
        info2 = []
        for an in range(alts):
            info1 = list(self.none['info'])
            for ind,(key,val) in enumerate(tokenized):
                try:
                    conv = self.converters['info'][key]
                except KeyError:
                    conv = no_converter
                try:
                    v = conv(val,an)
                except ValueError as exc:
                    if val=='' or val=='.':
                        v = None
                    else:
                        raise exc
                info1[self.order['info'][key]] = v
            info2.append(info1)

        return info2

    def get_infod(self, INFO, alts=1, parse_null=False):
        """Given INFO string and number of alt alleles returns a list
        of dicts with data corresponding to alleles"""
        tokenized = self.tokenize_info(INFO)
        info2 = []
        for an in range(alts):
            info1 = {}
            for ind,(key,val) in enumerate(tokenized):
                try:
                    conv = self.converters['info'][key]
                except KeyError:
                    conv = no_converter
                try:
                    v = conv(val,an)
                except ValueError as exc:
                    if val=='' or val=='.':
                        v = None
                    else:
                        raise exc
                info1[key] = v
            info2.append(info1)
        return info2

    def tokenize_sampdata(self,FORMAT,SAMPLES):
        def _maybe_split(key,val):
            if not self.dtype['format'][key]['number'] in [0,1]:
                return val.split(',')
            else:
                return val

        fmt = FORMAT.split(':')
        _samples = SAMPLES.split('\t')
        sampdata = [[(k,_maybe_split(k,v)) for k,v in zip(fmt,d.split(':'))]\
                     for d in _samples]
        return sampdata

    def get_sampdata(self,row,alts=1):
        """Given SAMPLE data string and number of alt alleles returns a list
        of lists of lists with data corresponding to alleles, then samples,
        then fields."""
        sampdata = [[list(self.none['format']) \
                       for j in range(len(self.sample_ind))] \
                           for i in range(alts)]
        _samples = self.tokenize_sampdata(row.FORMAT, row.SAMPLES)
        
        for an in range(alts):
            for sample,sn in self.sample_ind.items():
                _sampdata = list(self.none['format'])
                for key,val in _samples[sn]:
                    v = self.converters['format'][key](val,an)
                    _sampdata[self.order['format'][key]] = v
                sampdata[an][sn] = _sampdata
        return sampdata

    def get_dtype(self,hh):
        if not hh in ('format','info'):
            raise ValueError
        dtype = []
        for field,props in self.dtype[hh].items():
            if props['size']==1:
                dtype.append( (field,props['dtype']) )
            else:
                dtype.append( (field,props['dtype'],props['size']) )
        return dtype

class _BCFDataParser(_DataParser):
    def __init__(self,dtype,sample_ind):
        self.dtype = dtype
        self.converters = {'info':{},'format':{}}
        self.none = {'info':[],'format':[]}
        for hh in ['info','format']:
            for field,props in self.dtype[hh].items():
                tp = props['dtype']
                sz = props['size']
                num = props['number']
                self.converters[hh][field] = _make_converter_func(tp,num,convert=False)
                self.none[hh].append(None if sz<=1 else [None]*sz)
        self.converters['format']['GT'] = \
            lambda v,a: tuple([e==a for e in v])
        self.order = {
            'info':{f:ind for ind,f in enumerate(self.dtype['info'])},
            'format':{f:ind for ind,f in enumerate(self.dtype['format'])}
        }
        self.sample_ind = sample_ind

    def tokenize_info(self, INFO):
        return INFO.items()

    def tokenize_sampdata(self, FORMAT, SAMPLES):
        return [v.items() for v in SAMPLES.values()]
                    
class RowIterator:
    def iterate(self):
        cnt = 0
        for line in dropwhile(lambda l: l.startswith('#'),self.fh):
            # TODO maybe avoid stripping
            row = VCFRow(*line.strip().split('\t',maxsplit=9),
                         rnum=cnt)
            yield row
            cnt += 1
        self.close()

    def __init__(self,reader,fh,check_order):
        self.fh = fh
        self.check_order = check_order
        self.reader = reader
        if self.check_order:
            self.iterator = _check_VCF_order(self.iterate())
        else:
            self.iterator = self.iterate()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def close(self):
        self.fh.close()

class RowByChromIterator(RowIterator):
    def iterate(self):
        rows = super().iterate()
        if self.check_order:
            rows = _check_VCF_order(rows)
        for chrom,it in groupby(rows,key=lambda r: r.CHROM):
            yield chrom,it
        
    def __init__(self,*args,**kwds):
        super().__init__(*args,**kwds)
        
class BCFRowIterator(RowIterator):
    def __init__(self, *args, **kwds):
        super().__init__(*args, *kwds)

    def iterate(self):
        cnt = 0
        for rec in self.fh:
            row = VCFRow(
                rec.contig,rec.pos,rec.id,
                rec.ref, ','.join(rec.alts), rec.qual, ','.join(list(rec.filter)),
                rec.info, rec.format, rec.samples,
                rnum=cnt)
            yield row
            cnt += 1
        self.close()

class VrtIterator():
    def iterate(self):
        variants = self.reader._variants_from_rows(
                self.rows,self.factory,self.parse_info,self.samps)
        for vrt in variants:
            yield vrt
        self.close()
        
    def __init__(self,row_iterator,factory,parse_info,samps,
                 ensure_sorted=True):
        try:
            self.fh = row_iterator.fh
        except AttributeError:
            self.fh = None
        self.rows = row_iterator
        self.reader = row_iterator.reader
        self.factory = factory
        self.parse_info = parse_info
        self.samps = samps
        if ensure_sorted:
            self.iterator = _ensure_sorted(self.iterate())
        else:
            self.iterator = self.iterate()
        
    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def close(self):
        if self.fh is None:
            return
        self.fh.close() 
    
class VrtByChromIterator(VrtIterator):
    def iterate(self):
        vrt = super().iterate()
        for chrom,it in groupby(vrt,key=lambda r: r[-1].chrom):
            yield chrom,_ensure_sorted(it)

    def __init__(self, *args, **kwds):
        kwds['ensure_sorted'] = False
        super().__init__(*args, **kwds)

        
def _get_reader(file, reference=None):
    if isinstance(file, str) and file.endswith('.bcf'):
        return BCFReader(file, reference)
    else:
        return VCFReader(file, reference)



"""
Module :mod:`genomvar.variant` contains classes representing genomic
alterations.

The hierarchy of the classes used in the package is the following::

                      VariantBase
                      /      |   \\
     AmbigIndel <-Indel      V   |
       |     |      / \     MNP  V
       | ----+--- Del  Ins   |  Haplotype
       | |   |          |    V
       V V   |          V   SNP
    AmbigDel -----> AmbigIns

All variants have ``start`` and ``end`` attributes defining a range they
act on and can be searched overlap for.

To test whether a variant is instance of some type
:meth:`~VariantBase.is_instance` method can be used.  Variant equality
can be tested using :meth:`~VariantBase.edit_equal`.

Objects can be instantiated directly, e.g.::

    >>> vrt = variant.MNP('chr1',154678,'GT')
    >>> print(vrt)
    <MNP chr1:154678-154680 NN/GT>

This will create an MNP which substitutes positions 154678 and 154679 on
chromosome 1 for ``GT``.

Alternatively variants can be created using :class:`VariantFactory`
objects.  This class can work with VCF-like notation. For example, ::

    >>> fac = VariantFactory()
    >>> vrt = fac.from_edit('chr15',575,'TA','T')
    >>> print(vrt)
    <Del chr15:576-577 A/->

Position is **0-based** so it creates a deletion at position 577
of chromosome 15.

Alternatively, limited subset of HGVS notation is supported (numbering
in HGVS strings is **1-based** following the spec)::

    >>> vrt = fac.from_hgvs('chr1:g.15C>A')
    >>> print(vrt)
    <SNP chr1:14 C/A>

Variant sets defined in :mod:`genomvar.varset` use class :class:`GenomVariant`.
Objects of this class contain genomic alteration (attribute ``base``)
and optionally, genotype (attribute ``GT``) and other attributes
commonly found in VCF files (attribute ``attrib``).  Attribute
``base`` is an object of some ``VariantBase`` subclass (SNPs, Deletions
etc.).
"""
from collections import OrderedDict
import itertools
import re
import numpy as np
from rbi_tree.tree import ITree
from genomvar.utils import _strip_ref_alt
from genomvar import Reference,MAX_END,NoVCFNotationError,\
    StructuralVariantError
from genomvar.vcf_utils import (row_tmpl as vcf_row_template,
                                VCFRow, field_writer_simple)


hgvs_regex = {'SNP':'([0-9]+)([AGTC])>([AGTC])',
              'Del':'([0-9]+)(?:_([0-9]+))?del',
              'Ins':'[0-9]+_([0-9]+)ins([AGTC]+)',
              'DelIns':{0:'([0-9]+)_([0-9]+)delins([ATGC]+)',
                        1:'([0-9]+)delins([ATGC]+)'}}


class VariantBase(object):
    """
    Base class for the other genomic variant classes.
    """
    def __init__(self,chrom,start,end,ref,alt):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.ref = ref
        self.alt = alt

    __slots__ = 'chrom','start','end','ref','alt','_key'

    def __str__(self):
        return '<{} {}:{}-{} {}/{}>'\
            .format(type(self).__name__,self.chrom,self.start,self.end,
                    self.ref if self.ref else '-',
                    self.alt if self.alt else '-')
    def __repr__(self):
        return '{}(chrom="{}",start={},end={},ref="{}",alt="{}")'\
            .format(type(self).__name__,
                    self.chrom,self.start,self.end,
                    self.ref,self.alt)

    def edit_equal(self,other):
        """
        Returns True if ``self`` represents the same alteration as the ``other``
        """
        return self.key==other.key

    def get_key(self):
        return (type(self),self.chrom,self.start,self.end,self.alt)

    def __getattribute__(self,name):
        return object.__getattribute__(self,name)

    @classmethod
    def is_subclass(cls,other):
        """
        Checks is variant class is subclass of ``other``.  See module
        :mod:`genomvar.variant` documentaion for class hierarchy.
        """
        if issubclass(cls,other):
            return True
        else:
            return False

    def is_instance(self,cls):
        """Returns True if object's variant class is subclass of
        ``cls``.  See module :mod:`genomvar.variant` documentation for
        class hierarchy.
        """
        if issubclass(type(self),cls):
            return True
        else:
            return False

    def tolist(self):
        return [self.chrom,self.start,self.end,
                self.ref,self.alt,type(self),0,0]
        
    @property
    def vtp(self):
        """Variant class"""
        return type(self)
        
    @property
    def key(self):
        """Tuple representation of variant."""
        try:
            return self._key
        except AttributeError:
            self._key = self.get_key()
            return self._key

    @key.deleter
    def key(self,value):
        delattr(self,'_key')

    # def _format_row(self,pos,ref,alt,**kwds):
    #     row = VCFRow(self.chrom,pos,kwds.get('id', '.'),
    #                  ref.upper(),alt.upper(),
    #                  kwds.get('qual', '.'),
    #                  kwds.get('filter','.'),
    #                  kwds.get('info', '.'),
    #                  kwds.get('format'),
    #                  kwds.get('samples'))
    #     return row

    def _get_ref_notation(self,start,end, vcf_notation=None):
        # try:
        #     vcf_notation = self.attrib['vcf_notation']
        # except AttributeError:
        #     raise NoVCFNotationError
        # except KeyError:
        #     if not 'vcf_notation' in self.attrib:
        #         raise NoVCFNotationError
        start_ = start-vcf_notation['start']
        end_ = end-vcf_notation['start']
        if start_>=0 and end_<=len(vcf_notation['ref']):
            ref = vcf_notation['ref'][start_:end_]
        else:
            raise NoVCFNotationError('Out of bounds of vcf notation ref')
        return ref

    def _get_vcf_ref(self,start,end,vcf_notation=None,reference=None):
        if not vcf_notation is None:
            try:
                return self._get_ref_notation(
                    start,end, vcf_notation=vcf_notation)
            except NoVCFNotationError:
                if reference is None:
                    raise ValueError('No reference')

        if not reference is None:
            return reference.get(self.chrom, start, end)
        else:
            raise ValueError('Reference is required')

class MNP(VariantBase):
    """
    Multiple-nucleotide polymorphism.  Substitute N nucleotides of the
    reference for N other nucleotides.

    For instantiation it requires chromosome,position and 
    alternative sequence. ``end`` will inferred from ``start`` and ``alt``.
    ``ref`` is also optional.

    >>> from genomvar import variant
    >>> variant.MNP('chr1',154678,'GT')
    MNP("chr1",154678,"GT")
    
    """
    def __init__(self,chrom,start,alt,end=None,ref=None):
        if not end:
            end = start+len(alt)
        super().__init__(chrom,start,end,ref,alt)

    def __str__(self):
        return '<{} {}:{}-{} {}/{}>'\
            .format(type(self).__name__,self.chrom,self.start,self.end,
                    self.ref if self.ref else 'N'*(self.end-self.start),
                    self.alt if self.alt else '-')

    def __repr__(self):
        return 'MNP("{}",{},"{}")'\
            .format(self.chrom,self.start,self.alt)

    def nof_unit_vrt(self):
        return self.end-self.start

    def get_key(self):
        return ('MNP',self.chrom,self.start,self.end,self.alt)

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self.ref if self.ref else self._get_vcf_ref(
            self.start,self.end,vcf_notation=vcf_notation,reference=reference)
        alt = self.alt
        pos = self.start + 1 # to 1-based
        return pos, ref, alt
        
class SNP(MNP):
    """
    Single-nucleotide polymorphism.

    For instantiation it requires chromosome, position and 
    alternative sequence.

    >>> from genomvar import variant
    >>> variant.SNP('chr1',154678,'T')
    SNP("chr1",154678,"T")
    
    """
    def __init__(self,chrom,start,alt,end=None,ref=None):
        super().__init__(chrom=chrom,start=start,end=start+1,
                         ref=ref,alt=alt)
    def __repr__(self):
        return 'SNP("{}",{},"{}")'\
            .format(self.chrom,self.start,self.alt)

    def __str__(self):
        return '<{} {}:{} {}/{}>'\
            .format(type(self).__name__,self.chrom,self.start,
                    self.ref if self.ref else '-',
                    self.alt if self.alt else '-')

    def nof_unit_vrt(self):
        return 1

    def get_key(self):
        return ('SNP',self.chrom,self.start,self.alt)

class Indel(VariantBase):
    """Abstract class to accomodate insertion/deletion subclasses"""
    def nof_unit_vrt(self):
        return 1

    def ambig_equal(self,other):
        """Returns true if two variants are equal up to indel ambiguity."""
        return self.edit_equal(other) or self.shift_equal(other)


class Ins(Indel):
    """
    Insertion of nucleotides. For instantiation ``chrom``, 
    ``start``  and 
    inserted sequence (``alt``) are required.

    Start and end denote the nucleotide after the inserted sequence,
    i.e. ``start`` is 0-based number of a nucleotide after insertion,
    ``end`` is ``start+1`` by definition.
    
    >>> from genomvar.variant import Ins
    # Insertion of TA before position chr2:100543.
    >>> print(Ins('chr2',100543,'TA'))
    <Ins chr2:100543 -/TA>
    """
    def __init__(self,chrom,start,alt,end=None,ref=None):
        super().__init__(chrom,start,end=start+1,ref=ref,alt=alt)

    def __repr__(self):
        return 'Ins("{}",{},"{}")'\
            .format(self.chrom,self.start,self.alt)

    def __str__(self):
        return '<{} {}:{} -/{}>'\
            .format(type(self).__name__,self.chrom,self.start,
                    self.alt if self.alt else '-')

    def get_key(self):
        return ('Ins',self.chrom,self.start,self.alt)
    
    def shift_equal(self,other):
        if not other.is_instance(Ins):
            return False
        elif other.is_instance(AmbigIns):
            return other.shift_equal(self)
        else: # regular ins
            return self.edit_equal(other)

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self._get_vcf_ref(
            self.start-1,self.start,vcf_notation=vcf_notation,
            reference=reference)
        alt = ref + self.alt
        return self.start, ref, alt

class Del(Indel):
    """
    Deletion of nucleotides. For instantiation ``chrom``,
    ``start`` (0-based),
    and ``end`` (position ``end`` is excluded) are required.
    
    >>> from genomvar.variant import Del
    # Deletion of 3 nucleotides starting at chr3:7843488 (0-based)
    >>> print(Del('chr3',7843488,7843488+3))
    <Del chr3:7843488-7843491 NNN/->
    """
    def __init__(self,chrom,start,end,ref=None,alt=None):
        super().__init__(chrom,start,end,ref,alt)

    def __repr__(self):
        return 'Del("{}",{},{})'\
            .format(self.chrom,self.start,self.end)

    def __str__(self):
        return '<{} {}:{}-{} {}/->'\
            .format(type(self).__name__,self.chrom,self.start,self.end,
                    self.ref if self.ref else 'N'*(self.end-self.start))

    def get_key(self):
        return ('Del',self.chrom,self.start,self.end)
    
    def shift_equal(self,other):
        if not other.is_instance(Del):
            return False
        elif other.is_instance(AmbigDel):
            return other.shift_equal(self)
        else: # regular ins
            return self.edit_equal(other)

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self._get_vcf_ref(
            self.start-1, self.end,vcf_notation=vcf_notation,
            reference=reference)
        alt = ref[0]
        return self.start, ref, alt

class AmbigIndel(Indel):
    """Class representing indel which position is ambigous.  Ambiguity means the
    indel could be applied in any position of some region resulting in the same
    alternative sequence.
    """
    def __str__(self):
        return '<{} {}:{}-{}({}-{}) {}/{}>'\
            .format(type(self).__name__,
                    self.chrom,self.start,self.end,
                    self.act_start,self.act_end,
                    self.ref if self.ref else '-',
                    self.alt if self.alt else '-')

    # def shift_equal(self,other):
    #     if other.is_instance(AmbigIndel):
    #         a = (other.act_start-self.act_start) % len(self.seq)
    #         other_seq = other.seq
    #     elif not other.is_instance(Indel):
    #         return False
    #     else:
    #         a = (other.start-self.act_start) % len(self.seq)
    #         other_seq = other.alt if other.is_instance(Ins) else other.
    #         self_shift = self.seq[a:]+self.seq[:a]
    #         if self_shift==other.seq and self.chrom==other.chrom \
    #                  and self.vtp==other.vtp:
    #             return True
    #         else:
    #             return False
    #     else:
            


    def tolist(self):
        return [self.chrom,self.start,self.end,
                self.ref,self.alt,type(self),
                self.act_start,self.act_end]

class AmbigIns(AmbigIndel,Ins):
    """
    Class representing indel which position is ambigous.  Ambiguity
    means the indel could be applied in any position of some region
    resulting in the same alternative sequence.

    Let the reference file ``test.fasta`` contain a toy sequence::

      >seq1
      TTTAATA

    Consider a variant extending  3 ``T`` s in the beginning by one 
    more T. It can be done in several places so the corresponding
    insertion can be given as and ``AmbigIns`` object::

      >>> from genomvar import Reference
      >>> from genomvar.variant import VariantFactory
      >>> fac = VariantFactory(Reference('test.fasta'),normindel=True)
      >>> print( fac.from_edit('seq1',0,'T','TT') )
      <AmbigIns seq1:0-4(1-2) -/T>
    
    Positions 1 and 2 are actual start and end meaning that T
    is inserted before nucleotide located 1-2.
    Positions 0-4 indicate that start and end can be extended to 
    these values resulting in the same alteration.
    """
    def __init__(self,chrom,start,end,alt,ref=None):
        start,self.act_start = start
        end,self.act_end = end
        VariantBase.__init__(self,chrom=chrom,start=start,end=end,
                             ref=ref,alt=alt)
        
    # def edit_equal(self,other):
    #     return other.is_instance(Ins) \
    #         and self.chrom==other.chrom\
    #         and self.act_start==(other.act_start \
    #                   if other.is_instance(AmbigIns) else other.start)\
    #         and self.alt==other.alt

    @property
    def seq(self):
        return self.alt

    def __repr__(self):
        return 'AmbigIns("{}",start=({},{}),end=({},{}),alt="{}")'\
            .format(self.chrom,self.start,self.act_start,self.end,
                    self.act_end,self.alt)

    def get_key(self):
        return ('Ins',self.chrom,self.act_start,self.alt)

    def shift_equal(self,other):
        if not other.is_instance(Ins):
            return False
        elif other.is_instance(AmbigIns):
            a = (other.act_start-self.act_start) % len(self.seq)
            other_seq = other.seq
        else:
            a = (other.start-self.act_start) % len(self.seq)
            other_seq = other.alt
        self_shift = self.seq[a:] + self.seq[:a]
        if self_shift==other_seq and self.chrom==other.chrom:
            return True
        else:
            return False

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self._get_vcf_ref(
            self.start-1, self.start,
            vcf_notation=vcf_notation,reference=reference)
        a = (self.act_start-self.start) % len(self.seq)
        shifted = self.seq[-a:]+self.seq[:-a]
        alt = ref + shifted
        return self.start, ref, alt

class AmbigDel(AmbigIndel,Del):
    """
    Class representing del which position is ambigous.  Ambiguity means
    the same number of positions could be deleted in some range
    resulting in the same aternative sequence.

    Let the reference file ``test.fasta`` contain a toy sequence::

        >seq1
        TCTTTTTGACTGG

    >>> fac = VariantFactory(Reference('test.fasta'),normindel=True) >>>
    print( fac.from_edit('seq1',1,'CTTTTTGAC','C') ) <AmbigDel
    seq1:1-11(2-10) TTTTTGAC/->

    Deletion of TTTTTGAC starts at 2 and ends on 9th nucleotide
    (including 9th resulting in range 2-10).  1-11 denote that start and
    end can be extended to these values resulting in the same
    alteration.
    """
    def __init__(self,chrom,start,end,ref=None,alt=None):
        start,self.act_start = start
        end,self.act_end = end
        VariantBase.__init__(self,chrom=chrom,start=start,end=end,
                             ref=ref,alt=alt)

    def __repr__(self):
        return 'AmbigDel("{}",start=({},{}),end=({},{}))'\
            .format(self.chrom,self.start,self.act_start,
                    self.end,self.act_end)

    def get_key(self):
        return ('Del',self.chrom,self.act_start,self.act_end)

    @property
    def seq(self):
        return self.ref

    def shift_equal(self,other):
        if not other.is_instance(Del):
            return False
        elif other.is_instance(AmbigDel):
            other_span = other.act_end - other.act_start
            if other.start==self.start and other.end==self.end \
                       and other_span==self.act_end-self.act_start:
                return True
            else:
                return False
        else: # regular del
            other_span = other.end - other.start
            if other.start>=self.start and other.end<=self.end \
                        and other_span==self.act_end-self.act_start:
                return True
            else:
                return False

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self._get_vcf_ref(
            self.start-1, self.start+(self.act_end-self.act_start),
            vcf_notation=vcf_notation,
            reference=reference)
        alt = ref[0]
        return self.start, ref, alt

class Mixed(VariantBase):
    """Combination of Indel and MNP. Usage of this class is discouraged
    and exists for compatibility."""
    def __init__(self,chrom,start,end,alt,ref=None):
        super().__init__(chrom=chrom,start=start,end=end,ref=ref,alt=alt)

    def __repr__(self):
        return '{}("{}",{},{},alt="{}")'\
            .format(type(self).__name__,
                    self.chrom,self.start,self.end,
                    self.alt)

    def nof_unit_vrt(self):
        return min(len(self.alt),self.end-self.start) + 1

    def get_key(self):
        return ('Mixed',self.chrom,self.start,self.end,self.alt)

    def _get_vcf_notation(self,vcf_notation=None,reference=None):
        ref = self.ref if self.ref else self._get_vcf_ref(
            self.start,self.end,vcf_notation=vcf_notation,reference=reference)
        alt = self.alt
        pos = self.start + 1 # to 1-based
        return pos, ref, alt

class Haplotype(VariantBase):
    """
    An object representing genome variants on the
    same chromosome (or contig).

    Can be instantiated from a list of GenomVariant objects using
    :meth:`Haplotype.from_variants` class method.
    """
    def __init__(self,chrom,variants):
        start = min([v.start for v in variants])
        end = max([v.end for v in variants])

        super().__init__(chrom,start,ref='-',alt='-',end=end)
        self.data = ITree()
        ivl_id = [self.data.insert(v.start,v.end,v) for v in variants]

    def __str__(self):
        return '<Haplotype {}:{}-{} of {} variants>'\
            .format(self.chrom,self.start,
                    self.end,len(self._variants))
    def __repr__(self):
        return self.__str__()

    def nof_unit_vrt(self):
        return sum([v.nof_unit_vrt() for v in self.variants])

    def find_vrt(self,start=0,end=MAX_END):
        """
        Finds variants in specified region of a haplotype.
        
        Parameters
        ----------
        start : int 
            Start of search interval. *Default: 0*

        end: int
            End of search interval. *Defaults to MAX_END*

        yields
        -------
        vrt : variants
        """
        for s,e,v in self.data.find(start,end):
            yield v

    @classmethod
    def from_variants(cls,variants):
        """
        Create haplotype from a list of variants.

        Parameters
        ----------
        variants : list of variants
            Variants to instantiate haplotype from.

        Returns
        -------
        hap : Haplotype
            Haplotype variant object.
        """

        if len(variants)==0:
            raise ValueError('variants should be non-empty list')

        # infer chromosome
        chroms = set([o.chrom for o in variants])
        if len(chroms)>1:
            raise ValueError('variants on different chromosomes')
        elif len(chroms)==1:
            chrom = list(chroms)[0]
        else:
            raise ValueError('Chromosome not set')
        seen = {}
        for vrt in variants:
            if vrt.key in seen:
                raise ValueError('Non-unique variants {} and {}'\
                                 .format(vrt,seen[vrt.key]))
            else:
                seen[vrt.key] = vrt
        hap = cls.__new__(cls)
        hap.__init__(
            chrom,variants=sorted(
                variants,key=lambda v: (v.start,v.end)))
        return hap

    @property
    def variants(self):
        for s,e,v in self.data.iter_ivl():
            yield v

    def get_key(self):
        return tuple([v.key for v in self.variants])

class Null(VariantBase):
    def nof_unit_vrt(self):
        return 0

class Asterisk(VariantBase):
    def nof_unit_vrt(self):
        return 0

    def _get_vcf_notation(self,*args,**kwds):
        raise NotImplementedError

class GenomVariant(object):
    """
    This is a variant class holding gemomic alteration 
    (a VariantBase subclass object)
    and extra attributes. Underlying variant is accessible as ``base`` attribute.
    On top it has genotype as ``GT`` attribute and ``attrib`` containing
    extra attributes parsed from VCF files.  All attributes of
    underlying ``base`` (start, end etc.) are accessible from
    GenomVariant object.
    """
    def __init__(self,base,GT=None,attrib=None):
        if not isinstance(base,VariantBase):
            raise TypeError('Should be VariantBase or str')
        self.base = base
        self.GT = GT
        self.attrib = attrib if attrib else {}

    def __getattr__(self,name):
        if name=='base':
            raise AttributeError
        return object.__getattribute__(self.base,name)

    def __str__(self):
        return '<GenomVariant: {}'.\
            format(self.base.__str__()[1:],
                   getattr(self,'GT',None))

    def __repr__(self):
        return 'GenomVariant({})'.format(repr(self.base),str(self.GT))


    def edit_equal(self,other):
        """
        Check if GenomVariant holds the same genome alteration
        as ``other``.
        
        Parameters
        ----------
        variant : GenomVariant or any VariantBase
            variant to compare
        Returns
        -------
        equality : bool
            True if the same genomic alteration.
        """
        
        return self.base.edit_equal(other)

    def _get_ref_notation(self, start, end):
        try:
            vcf_notation = self.attrib['vcf_notation']
        except AttributeError:
            raise NoVCFNotationError
        except KeyError:
            if not 'vcf_notation' in self.attrib:
                raise NoVCFNotationError
        start_ = start-vcf_notation['start']
        end_ = end-vcf_notation['start']
        if start_>=0 and end_<=len(vcf_notation['ref']):
            ref = vcf_notation['ref'][start_:end_]
        else:
            raise NoVCFNotationError('Out of bounds of vcf notation ref')
        return ref

    def _get_vcf_ref(self,start,end,reference=None):
        try:
            ref = self._get_ref_notation(start,end)
        except NoVCFNotationError:
            if reference is None:
                raise ValueError('No reference')
            else:
                return reference.get(self.chrom, start, end)

class VariantFactory(object):
    """
    Factory class used to create Variant objects.  Can be instantiated
    without arguments or an instance of :class:`genomvar.Reference` can
    be given. Additionnaly if ``normindel=True`` (reference should be
    given) indels will be checked for ambiguity, and if ambigous instance
    of :class:`genomvar.variant.AmbigIndel` will be returned.

    >>> from genomvar import Reference
    >>> from genomvar.variant import VariantFactory
    >>> reference = Reference('test.fasta')
    >>> fac = VariantFactory(reference,normindel=True)
    """
    def __init__(self,reference=None,normindel=False):
        if isinstance(reference,Reference):
            self.reference = reference
        elif isinstance(reference,str):
            self.reference = Reference(reference)
        elif reference is None:
            self.reference = None
        else:
            raise ValueError('reference not understood')
        if normindel:
            if self.reference is None:
                raise ValueError('reference needed for indel normalization')
        self.normindel = normindel

    def from_hgvs(self,st):
        """Returns a variant from a HGVS notation string. Only some of
        possible HGVS notations are supported as listed below::

          >>> fac = VariantFactory()
          >>> print(fac.from_hgvs('chr1:g.15C>A'))
          <SNP chr1:14 C/A>
          >>> print(fac.from_hgvs('chrW:g.19_21del'))
          <Del chrW:18-21 NNN/->
          >>> print(fac.from_hgvs('chr24:g.10_11insCCT'))
          <Ins chr24:10 -/CCT>
          >>> print(fac.from_hgvs('chr24:g.10delinsGA'))
          <Mixed chr24:9-10 -/GA>
          >>> print(fac.from_hgvs('chr23:g.145_147delinsTGG'))
          <MNP chr23:144-147 NNN/TGG>
        """
        chrom,rest = st.split(':',maxsplit=1)
        tp,posedit = rest.split('.',maxsplit=1)
        if '>' in posedit:
            match = re.match(hgvs_regex['SNP'],posedit)
            pos,ref,alt = match.groups()
            start = int(pos)-1
            return SNP(chrom=chrom,start=start,alt=alt,ref=ref)
        elif posedit.endswith('del'):
            match = re.match(hgvs_regex['Del'],posedit)
            pos1,pos2 = match.groups()
            start = int(pos1)-1
            end = int(pos2) if pos2 else start+1
            if not self.normindel:
                return Del(chrom=chrom,start=start,end=end)
            else:
                start2,end2,ref,alt,cls = \
                      self._maybe_norm_del(chrom=chrom,start=start,end=end,
                                ref=self.reference.get(chrom,start,end))
                return cls(chrom,start2,end2,ref,alt)
        elif 'delins' in posedit:
            m = re.match(hgvs_regex['DelIns'][0],posedit)
            if m:
                pos1,pos2,ins = m.groups()
                start = int(pos1)-1
                end = int(pos2)
            else:
                m2 = re.match(hgvs_regex['DelIns'][1],posedit)
                if m2:
                    pos,ins = m2.groups()
                    start = int(pos)-1
                    end = start+1
                else:
                    raise ValueError('Unable to parse '+st)
            if end-start==len(ins): # MNP
                return MNP(chrom=chrom,start=start,end=end,alt=ins)
            else: #Mixed
                return Mixed(chrom=chrom,start=start,end=end,alt=ins)
        elif 'ins' in posedit:
            match = re.match(hgvs_regex['Ins'],posedit)
            pos,alt = match.groups()
            start = int(pos)-1
            if not self.normindel:
                return Ins(chrom,start=start,alt=alt)
            else:
                start2,end2,ref,alt,cls = \
                      self._maybe_norm_ins(chrom=chrom,start=start,
                                             alt=alt)
                return cls(chrom,start2,end2,ref,alt)
        else:
            raise ValueError('Unable to parse '+st)

    def from_edit(self,chrom,start,ref,alt):
        """
        Takes chrom, start, ref and alt (similar to a row in VCF file,
        but start is 0-based) and returns a variant object.

          >>> vrt = fac.from_edit('chr15',575,'TA','T')
          >>> print(vrt)
          <Del chr15:576-577 A/->
        
        Method attempts to strip ref and alt sequences from left and right
        (in this order), for example,
        
          >>> vrt = fac.from_edit('chr15',start=65,ref='CTG',alt='CTC')
          >>> print(vrt)
          <SNP chr15:67-68 G/C>
        """
        # print('!! processing', chrom, start, ref, alt)
        chrom,start,end,ref,alt,cls = \
            self._parse_edit(chrom,start,ref,alt)
        return cls(chrom=chrom,start=start,end=end,ref=ref,alt=alt)
    
    def _parse_edit(self,chrom,start,ref,alt):
        """
        Given a chrom, start, ref and alt record it returns a tuple
        for variant instantiation.
        """
        # store the start for later use
        start0 = start + min(len(ref),len(alt))

        if alt=='*':
            return (chrom,start,start+len(ref),ref,alt,Asterisk)

        if len(ref)==1 and len(alt)==1 and ref!=alt:
            tp=SNP
            end = start+len(ref)
            return (chrom,start,end,ref,alt,tp)
        else:
            ref,alt,shift = _strip_ref_alt(ref,alt)
            start += shift
            if not ref: # Ins
                if '[' in alt or ']' in alt:
                    raise StructuralVariantError
                if not self.normindel:
                    return (chrom,start,start+1,'',alt,Ins)
                else:
                    start2,end2,ref,alt,cls = \
                          self._maybe_norm_ins(chrom=chrom,start=start,
                                            alt=alt)
                    return (chrom,start2,end2,ref,alt,cls)
            elif not alt: #Del
                if not self.normindel:
                    return (chrom,start,start+len(ref),ref,'',Del)
                else:
                    start2,end2,ref,alt,cls = \
                          self._maybe_norm_del(chrom=chrom,start=start,
                                               end=start+len(ref),ref=ref)
                    return (chrom,start2,end2,ref,alt,cls)
                
            elif len(ref)==len(alt): #MNP
                end = start+len(ref)
                tp = MNP if len(ref)>1 else SNP
                return (chrom,start,end,ref,alt,tp)
            elif '<' in alt:
                raise StructuralVariantError
            else: #Mixed
                return (chrom,start,start+len(ref),ref,alt,Mixed)

    def _maybe_norm_ins(self,chrom,start,alt,max_len=MAX_END):
        """Check insertion for normalization"""
        left_norm = False
        right_norm = False
        # _start will be the first position to check for
        # equality with the unit
        _start = start
        it = itertools.cycle(reversed(alt))
        while _start >= 1:
            _start -= 1
            rl = str(self.reference.get(chrom,_start,_start+1))
            ul = next(it)
            if rl != ul:
                _start += 1
                break
            left_norm = True

        # Checking end
        it = itertools.cycle(alt)
        _end = start
        while _end < max_len:
            rl = str(self.reference.get(chrom,_end,_end+1))
            ul = next(it)
            if rl != ul:
                _end += 1
                break
            _end += 1
            right_norm = True

        if left_norm or right_norm:
            return ((_start,start),(_end,start+1),'',alt,AmbigIns)
        else:
            return (_start,_start+1,'',alt,Ins)

    def _maybe_norm_del(self,chrom,start,end,ref=None,max_len=MAX_END):
        """Check deletion if needs to be normalized"""

        if not ref:
            ref = self.reference(chrom,start,end)
        left_norm = False
        right_norm = False
        # _start will be the first position to check for
        # equality with the unit
        _start = start
        it = itertools.cycle(reversed(ref))
        while _start >= 0:
            _start -= 1
            rl = str(self.reference.get(chrom,_start,_start+1))
            ul = next(it)
            if rl != ul:
                _start += 1
                break
            left_norm = True

        # Checking end
        it = itertools.cycle(ref)
        _end = start + len(ref)
        while _end < max_len:
            rl = str(self.reference.get(chrom,_end,_end+1))
            ul = next(it)
            if rl != ul:
                break
            _end += 1
            right_norm = True

        if left_norm or right_norm:
            return ((_start,start),(_end,end),ref,'',AmbigDel)
        else:
            return (_start,_end,ref,'',Del)



import os
from collections import namedtuple
from jinja2 import Environment,FileSystemLoader
import numpy as np

VCF_fields = ["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER",  "INFO",
              "FORMAT",  "SAMPLES"]
VCF_info_fields = ["NAME",  "NUMBER",  "TYPE",  "DESCRIPTION",  "SOURCE",  "VERSION"]
VCFInfoSpec = namedtuple('VCFInfoSpec', VCF_info_fields)
# Map of VCF types to NumPy types
string2dtype = {'Float':np.float64,'Integer':np.int64,
                'String':np.object_,'Flag':np.bool_,
                'Character':'U1'}
# inverse for writing VCFs
dtype2string = {v:k for k,v in string2dtype.items()}


class VCFRow(object):
    """Tuple-like object storing variant information in VCF-like form.

    str() returns a string, formatted as a row in VCF file."""
    def __init__(self,CHROM,POS,ID,REF,ALT,QUAL,FILTER,INFO,
                 FORMAT=None,SAMPLES=None,rnum=None):
        self.CHROM = CHROM
        self.POS = int(POS)
        self.ID = ID
        self.REF = REF
        self.ALT = ALT
        self.QUAL = QUAL
        self.FILTER = FILTER
        self.INFO = INFO
        self.FORMAT = FORMAT
        self.SAMPLES = SAMPLES
        self.rnum = rnum

    __slots__ = [*VCF_fields, 'rnum']

    @staticmethod
    def _to_string(v):
        if v is None:
            return '.'
        else:
            return str(v)

    def __repr__(self):
        return '<VCFRow {}:{} {}->{}>'\
            .format(self.CHROM,self.POS,self.REF,self.ALT)
    def __str__(self):
        fields = [self._to_string(getattr(self, a)) for a in VCF_fields[:8]]
        if not self.FORMAT is None:
            fields += [self._to_string(self.FORMAT),str(self.SAMPLES)]
        return '\t'.join(fields)

tmpl_dir = os.path.join(os.path.dirname(__file__),'tmpl')
env = Environment(
    loader=FileSystemLoader(tmpl_dir),
)
header_simple = env.get_template('vcf_head_simple.tmpl')
header = env.get_template('vcf_head.tmpl')
row_tmpl = env.get_template('vcf_row.tmpl')

def _make_field_writer_func(tp, num):
    """tp can be a string in VCF notation or internal NumPy types."""
    if isinstance(tp, str):
        try:
            tp = string2dtype[tp]
        except KeyError:
            raise ValueError(
                'Allowed types "Type" in INFO spec are {} while found {}'\
                .format(','.join(string2dtype), tp))
    if tp==np.bool_:
        return lambda k,v: str(k)
    if tp==np.int_:
        tostring = lambda v: str(int(v)) if not np.isnan(v) else '.'
    else:
        tostring = lambda v: str(v) if not v is None else '.'
    if num in ('.', 'G') or (isinstance(num, int) and num>1):
        return lambda k,v: '{}={}'.format(
            k, ','.join(map(tostring, v)) \
            if not v is None else '.')
    elif num in (1, 'A'): # Alleles are split on read-in
        return lambda k,v: '{}={}'.format(
            k, tostring(v) if not v is None else '.')
    else:
        allowed = 'A, G, . or and integer'
        raise ValueError(
            'Allowed types for "Number" INFO spec are {} while found {}'\
            .format(allowed, num))

def issequence(seq):
    """
    Is seq a sequence (ndarray, list or tuple)?

    """
    return isinstance(seq, (np.ndarray, tuple, list))
    
def field_writer_simple(k,v):
    if issequence(v):
        res = '{}={}'.format(k, ','.join(map(str, v)))
    else:
        res = '{}={}'.format(k, str(v))
    return res



import itertools
import warnings
from genomvar.vcf import VCFReader, VCFWriter,\
    dtype2string,_isindexed
from genomvar.vcf_utils import row_tmpl, header as vcf_header
from genomvar.varset import VariantFileSet,IndexedVariantFileSet
from genomvar import variant

def _same_order(chroms1,chroms2):
    common_chroms = set(chroms1).intersection(chroms2)
    _chroms1 = list(filter(lambda e: e in common_chroms,chroms1))
    _chroms2 = list(filter(lambda e: e in common_chroms,chroms2))
    if _chroms1==_chroms2:
        return True
    else:
        return False

def _cmp_vcf(f1,f2,out,match_partial=False,chunk=1000):
    """
    Writes comparison of two VCF files to a specified file handle.
    """
    info = [('VT', 1, 'String', 'Variant type'),
            ('whichVCF', 1, 'String', 'Which input VCF contains the variant; first, second or both'),
            ('ln', 1, 'Integer', 'Line number in input VCF variant originating from'),
            ('ln2', '.', 'Integer', 'If whichVCF is both indicates line numberin the second file')]

    writer = VCFWriter(info_spec=info)
    header = writer.get_header()
    out.write(header)

    if _isindexed(f1):
        vs1 = IndexedVariantFileSet(f1)
    else:
        warnings.warn('{} not indexed; may impact performance.'.format(f1))
        vs1 = VariantFileSet(f1)
    if _isindexed(f2):
        vs2 = IndexedVariantFileSet(f2)
    else:
        warnings.warn('{} not indexed; may impact performance.'.format(f2))
        vs2 = VariantFileSet(f2)

    _which = {0:'first',1:'second',2:'both'}
    nof_vrt = {i:0 for i in _which}
    cb = lambda m: [v.attrib['vcf_notation']['row'] for v in m]
    for which,vrt in vs1._cmp_vrt(vs2,action='all')\
                        .iter_vrt(callback=cb):
        
        nof_vrt[which] += vrt.nof_unit_vrt()
        if which==0:
            lineno = vrt.attrib['vcf_notation']['row']+vs1._reader.header_len+1
        elif which==1:
            lineno = vrt.attrib['vcf_notation']['row']+vs2._reader.header_len+1
        if which==2:
            lineno = vrt.attrib['vcf_notation']['row']+vs1._reader.header_len+1
            lineno2 = [vs2._reader.header_len+n+1 for n in vrt.attrib['cmp']]

        vrt.attrib['info'] = {'whichVCF':_which[which],
                              'ln':lineno,
                              'ln2':lineno2 if which==2\
                                  else None}
        try:
            row = writer.get_row(vrt)
        except ValueError as exc:
            if vrt.is_instance(variant.Haplotype) \
                    or vrt.is_instance(variant.Asterisk):
                continue
            else:
                raise exc
            
        out.write(str(row)+'\n')
    return nof_vrt



from setuptools import setup
import re

def readme():
    with open('README.rst') as f:
        desc = f.read()
        desc = re.sub(':class:`~?(.+)`',r'\1',desc)
        desc = re.sub(':meth:`~?(.+)`',r'\1',desc)
        return desc

setup(name='genomvar',
      version='0.2.14',
      description='Sequence variant analysis in Python',
      long_description=readme(),
      classifiers=[
          'Intended Audience :: Science/Research',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Software Development :: Libraries'],
      keywords= ['Bioinformatics', 'Computational Biology'],
      url='http://relativity.nprog.local/mikpom/genomvar',
      author='Mikhail Pomaznoy',
      author_email='mikpom@mailbox.org',
      license='BSD',
      packages=['genomvar'],
      package_data={
          'genomvar':['tmpl/*']
      },
      install_requires=[
          'rbi-tree>=0.1.3.1',
          'jinja2'],
      test_suite='unittest',
      tests_require=['setuptools'],
      zip_safe=False)
